Package: cooccure
Title: Co-Occurrence Network Construction and Manipulation
Version: 0.1.1
Authors@R: c(
    person("Mohammed", "Saqr", email = "saqr@saqr.me",
           role = c("aut", "cre", "cph")),
    person("Sonsoles", "López-Pernas", email = "sonsoles.lopez@uef.fi",
           role = c("aut", "cph")),
    person("Kamila", "Misiejuk", role = c("aut", "cph")))
Description: Constructs co-occurrence networks from several types of input
    data, such as delimited fields, long/bipartite tables, binary
    matrices, or wide sequences. Returns tidy edge data frames and
    supports optional scaling, splitting into several networks,
    thresholding, and subsetting. Provides eight similarity measures,
    including Jaccard, cosine, and association strength. Supports export
    to several network and file formats. Network construction and
    analysis methods follow Saqr, Lopez-Pernas, Conde, and
    Hernandez-Garcia (2024, <doi:10.1007/978-3-031-54464-4_15>).
License: MIT + file LICENSE
Encoding: UTF-8
Language: en-US
LazyData: true
RoxygenNote: 7.3.3
Imports: graphics, Matrix, methods, stats, utils
Suggests: igraph, cograph, Nestimate, tidygraph, testthat (>= 3.0.0),
        knitr, rmarkdown, shiny, DT
Config/testthat/edition: 3
VignetteBuilder: knitr
Depends: R (>= 3.5.0)
URL: https://github.com/mohsaqr/cooccure
BugReports: https://github.com/mohsaqr/cooccure/issues
NeedsCompilation: no
Packaged: 2026-04-22 20:37:56 UTC; mohammedsaqr
Author: Mohammed Saqr [aut, cre, cph],
  Sonsoles López-Pernas [aut, cph],
  Kamila Misiejuk [aut, cph]
Maintainer: Mohammed Saqr <saqr@saqr.me>
Repository: CRAN
Date/Publication: 2026-04-24 20:20:10 UTC
Built: R 4.5.2; ; 2026-04-24 23:25:04 UTC; unix

Package: cooccure
Title: Co-Occurrence Network Construction and Manipulation
Version: 0.1.1
Authors@R: c(
    person("Mohammed", "Saqr", email = "saqr@saqr.me",
           role = c("aut", "cre", "cph")),
    person("Sonsoles", "López-Pernas", email = "sonsoles.lopez@uef.fi",
           role = c("aut", "cph")),
    person("Kamila", "Misiejuk", role = c("aut", "cph")))
Description: Constructs co-occurrence networks from several types of input
    data, such as delimited fields, long/bipartite tables, binary
    matrices, or wide sequences. Returns tidy edge data frames and
    supports optional scaling, splitting into several networks,
    thresholding, and subsetting. Provides eight similarity measures,
    including Jaccard, cosine, and association strength. Supports export
    to several network and file formats. Network construction and
    analysis methods follow Saqr, Lopez-Pernas, Conde, and
    Hernandez-Garcia (2024, <doi:10.1007/978-3-031-54464-4_15>).
License: MIT + file LICENSE
Encoding: UTF-8
Language: en-US
LazyData: true
RoxygenNote: 7.3.3
Imports: graphics, Matrix, methods, stats, utils
Suggests: igraph, cograph, Nestimate, tidygraph, testthat (>= 3.0.0),
        knitr, rmarkdown, shiny, DT
Config/testthat/edition: 3
VignetteBuilder: knitr
Depends: R (>= 3.5.0)
URL: https://github.com/mohsaqr/cooccure
BugReports: https://github.com/mohsaqr/cooccure/issues
NeedsCompilation: no
Packaged: 2026-04-22 20:37:56 UTC; mohammedsaqr
Author: Mohammed Saqr [aut, cre, cph],
  Sonsoles López-Pernas [aut, cph],
  Kamila Misiejuk [aut, cph]
Maintainer: Mohammed Saqr <saqr@saqr.me>
Repository: CRAN
Date/Publication: 2026-04-24 20:20:10 UTC
Built: R 4.6.0; ; 2026-04-27 08:26:05 UTC; unix

# Generated by roxygen2: do not edit by hand

S3method(as_cograph,cooccurrence)
S3method(as_igraph,cooccurrence)
S3method(as_matrix,cooccurrence)
S3method(as_netobject,cooccurrence)
S3method(as_tidygraph,cooccurrence)
S3method(plot,cooccurrence)
S3method(print,cooccurrence)
S3method(summary,cooccurrence)
export(as_cograph)
export(as_igraph)
export(as_matrix)
export(as_netobject)
export(as_tidygraph)
export(co)
export(cooccurrence)
export(launch_app)

# cooccure <img src="man/figures/logo.png" align="right" height="200"  alt="" />

The **cooccure** R package enables building co-occurrence networks from multiple data formats. It accepts six input formats and supports multiple similarity measures, scaling methods, fractional counting, group-level splitting, and flexible filtering. Results are returned as a tidy edge data frame (`from`, `to`, `weight`, `count`) convertible to **igraph**, **tidygraph**, **cograph**, and **Nestimate** objects.

The main function `cooccurrence()` is also available as the short alias `co()`.

## Installation

```r
# CRAN release
install.packages("cooccure")

# Development version
remotes::install_github("mohsaqr/cooccure")
```

## Input formats

**cooccure** auto-detects the input format from the arguments provided. Six formats are supported (delimited field, multi-column delimited, long/bipartite, binary matrix, wide sequence, and list of character vectors), covering the most common shapes data comes in.

### 1. Delimited field

A *delimited field* is a single column where multiple items are stored as one string, separated by a consistent character such as `;`, `,`, `|`, or a `space`. This is the most common format in bibliometrics and text analysis, where each row represents a document.

Use the `field` argument to specify the column with the relevant values, and `sep` to specify the delimiter.

```r
df <- data.frame(
  id = 1:3,
  keywords = c("network; graph; matrix",
               "graph; algebra",
               "network; algebra; graph")
)
cooccurrence(df, field = "keywords", sep = ";")
```

Whitespace around the separator is automatically trimmed (`" network "` becomes `"network"`). Empty strings and NAs are dropped. Duplicate items within a row are de-duplicated.

### 2. Multi-column delimited

A *multi-column delimited* format is used when items are spread across multiple columns - for example, author keywords and index keywords in a Scopus export, or authors and affiliations. Values from all specified columns are pooled per row.

Use the `field` argument to specify the columns with the relevant values, and `sep` to specify the delimiter.

```r
df <- data.frame(
  author_kw = c("machine learning; nlp", "deep learning", "nlp"),
  index_kw  = c("classification", "image recognition", "text mining")
)
cooccurrence(df, field = c("author_kw", "index_kw"), sep = ";")
```

### 3. Long / bipartite

A *long* or *bipartite* format has one row per item-document pair, common in relational databases, survey data, and tidy data pipelines.

Use the `field` argument to specify the column containing the items, and the `by` argument to specify which column groups them into transactions.

```r
citations <- data.frame(
  paper_id = c(1, 1, 1, 2, 2, 3, 3, 3),
  reference = c("Smith2020", "Jones2019", "Lee2021",
                "Jones2019", "Lee2021",
                "Smith2020", "Lee2021", "Park2022")
)
cooccurrence(citations, field = "reference", by = "paper_id")
```

Use the `weight_by` argument to pass a numeric weight column for *weighted long* format — for example, LDA topic-document probabilities where each document contributes a topic with a given probability:

```r
theta <- data.frame(
  doc   = c("d1","d1","d1","d2","d2","d3","d3"),
  topic = c("T1","T2","T3","T1","T3","T2","T3"),
  prob  = c(0.6, 0.3, 0.1, 0.4, 0.6, 0.5, 0.5)
)
cooccurrence(theta, field = "topic", by = "doc", weight_by = "prob")
```

In *weighted long* format, the co-occurrence between items `i` and `j` is computed as `sum_d w_id * w_jd` (the sum of the products of their weights across all shared transactions) rather than a simple binary count. The `count` column still reports the number of transactions where both items appear together.

### 4. Binary matrix

A *binary matrix* is a document-term matrix where columns are items and values are 0/1 (absence or presence). This format is auto-detected when all values are 0 or 1 and no `field`, `by`, or `sep` arguments are provided.

```r
dtm <- matrix(c(1,1,0,1,
                0,1,1,0,
                1,0,1,1), nrow = 3, byrow = TRUE,
              dimnames = list(NULL, c("network", "graph", "algebra", "matrix")))
cooccurrence(dtm)
```

Works with both `matrix` and `data.frame` inputs. Columns without names are auto-named `V1`, `V2`, etc.

### 5. Wide sequence

A *wide sequence* format is used for non-binary data frames or matrices where each row is a sequence or record, and the unique values in each row form a transaction. This is the native format for sequence analysis tools like **TraMineR** and **tna**. Pass `field = "all"` to treat every column as a time point.

```r
sequences <- data.frame(
  t1 = c("A", "B", "A"),
  t2 = c("B", "C", "C"),
  t3 = c("C", NA,  NA)
)
cooccurrence(sequences, field = "all")
```

NAs, empty strings, and **TraMineR** void markers (`%`, `*`) are automatically removed.

### 6. List of character vectors

A *list of character vectors* is the most direct format, where each list element is a transaction containing a set of categorical items.

```r
baskets <- list(
  c("bread", "milk", "eggs"),
  c("bread", "butter"),
  c("milk", "eggs", "butter"),
  c("bread", "milk", "eggs", "butter")
)
cooccurrence(baskets)
```

## Similarity measures

The `similarity` argument controls how raw co-occurrence counts are normalized into a similarity or association measure. All similarity measures are based on two inputs: the co-occurrence count between two items ($C_{ij}$), and how frequently each item appears individually across all transactions ($f_i$, $f_j$).

```r
# Jaccard similarity
cooccurrence(baskets, similarity = "jaccard")

# Association strength
cooccurrence(papers, field = "keywords", sep = ";", similarity = "association")
```

### Which similarity to use?

- **Exploratory work**: Start with `"none"` to see raw counts and understand the data, then try `"jaccard"` or `"cosine"` for a balanced view.

- **General purpose**: `"jaccard"` is a good default choice. It normalizes co-occurrences by the union of transactions containing either item, applying a balanced penalty for non-overlap.

- **Bibliometric and scientometric networks**: `"association"` is recommended by van Eck & Waltman (2009) because it correctly accounts for the expected number of co-occurrences under independence. Two items that are both very frequent will naturally co-occur often; association strength discounts this, revealing which pairs co-occur *more than chance alone would predict*.

- **Detecting hierarchical/subset structure**: `"inclusion"` (Simpson coefficient) reveals when one item almost always appears with another, useful for finding items that are subsets of broader categories or dependency relationships.

- **Binary presence/absence networks**: `"dice"` when you only care *whether* items co-occur, not *how often*. Less strict than `"jaccard"`, it applies a less severe penalty for partial overlap.

- **Scale-invariant comparison**: `"cosine"` is invariant to absolute frequency, useful when comparing co-occurrence patterns across datasets of different sizes or when frequent items should not dominate the network.

- **Strict filtering**: `"equivalence"` (cosine squared) amplifies differences, pushing pairs with weak overlap closer to zero and retaining only the strongest associations.

- **Asymmetric tendencies**: `"relative"` normalizes each row so that edge weights sum to 1, capturing the relative tendency of one item to appear with another rather than absolute co-occurrence counts.

### Similarity method overview

| Method | Formula | Description | Best for |
|--------|---------|-------------|----------|
| `"none"` | $C_{ij}$ | Raw co-occurrence count | Exploratory analysis |
| `"jaccard"` | $\frac{C_{ij}}{f_i + f_j - C_{ij}}$ | Divides co-occurrences by the total number of transactions containing either item | General purpose |
| `"cosine"` | $\frac{C_{ij}}{\sqrt{f_i \cdot f_j}}$ | Divides co-occurrences by the geometric mean of item frequencies (Salton's cosine) | Scale-invariant comparison |
| `"inclusion"` | $\frac{C_{ij}}{\min(f_i, f_j)}$ | Divides co-occurrences by the frequency of the rarer item (Simpson coefficient) | Subset and hierarchical relationships |
| `"association"` | $\frac{C_{ij}}{f_i \cdot f_j}$ | Divides co-occurrences by the product of item frequencies, discounting chance co-occurrences (van Eck & Waltman, 2009) | Bibliometric networks |
| `"dice"` | $\frac{2 C_{ij}}{f_i + f_j}$ | Divides co-occurrences by the arithmetic mean of item frequencies | Binary presence/absence networks |
| `"equivalence"` | $\frac{C_{ij}^2}{f_i \cdot f_j}$ | Squares the co-occurrence count before dividing by the product of frequencies (cosine squared) | Strict filtering |
| `"relative"` | Row-normalized (each row sums to 1) | Normalizes each row so that all edge weights from an item sum to 1 | Asymmetric tendencies |

## Counting

The `counting` argument controls how much each transaction contributes to the co-occurrence count. Under full counting (default), each co-occurring pair adds 1 regardless of how many items are in the transaction. Under fractional counting, each pair adds $1/(n-1)$ where $n$ is the number of items in the transaction, preventing large transactions from dominating the network. For example, a document with 10 keywords creates 45 pairs under full counting but contributes only 1/9 per pair under fractional counting.

```r
# Full counting (default): each co-occurring pair adds 1
co(data, field = "keywords", sep = ";")

# Fractional: each pair adds 1/(n-1) where n = items in the transaction
co(data, field = "keywords", sep = ";", counting = "fractional")
```

## Scaling

The `scale` argument applies a transformation to the weights after similarity normalization, useful for visualization, thresholding, or feeding into downstream models.

```r
# Log-scaled Jaccard similarity
cooccurrence(baskets, similarity = "jaccard", scale = "log")

# Min-max scaled for visualization
cooccurrence(baskets, similarity = "cosine", scale = "minmax")
```

### Scaling method overview

| Method | Transformation | Description | Use case |
|--------|---------------|-------------|----------|
| `"minmax"` | Scale to $[0, 1]$ | Rescales all weights to the range $[0, 1]$ | Visualization and cross-network comparison |
| `"log"` | $\log(1 + w)$ | Applies a natural log transformation | Compressing heavy-tailed distributions |
| `"log10"` | $\log_{10}(1 + w)$ | Same as log but base 10 | When base 10 interpretation is preferred |
| `"sqrt"` | $\sqrt{w}$ | Square root transformation | Mild compression of skewed weights |
| `"binary"` | 1 if $w > 0$, else 0 | Converts all positive weights to 1 | Presence/absence networks |
| `"zscore"` | $(w - \mu) / \sigma$ | Standardizes weights to mean 0 and standard deviation 1 | Statistical comparison across networks |
| `"proportion"` | $w / \sum w$ | Divides each weight by the total sum of weights | Expressing edges as relative importance |

## Filtering

Three filtering arguments control which edges appear in the result:

- `min_occur`: drops any entity appearing in fewer than a specified number of transactions, removing rare items before co-occurrences are computed.
- `threshold`: keeps only edges with a weight at or above a specified value, applied after similarity normalization and scaling.
- `top_n`: keeps only the $n$ strongest edges by weight.

```r
# Drop entities appearing in fewer than 3 transactions
cooccurrence(baskets, min_occur = 3)

# Keep only edges with weight >= 0.5 (applied after similarity + scaling)
cooccurrence(baskets, similarity = "jaccard", threshold = 0.5)

# Keep only the 10 strongest edges
cooccurrence(baskets, top_n = 10)
```

All three can be combined for fine-grained control over the network size and density:

```r
cooccurrence(papers, field = "keywords", sep = ";",
             similarity = "association", min_occur = 2,
             threshold = 0.01, top_n = 50)
```

## Splitting by groups

The `split_by` argument computes a separate co-occurrence network for each level of a grouping variable and returns them in a single data frame with a `group` column. This is useful for comparing co-occurrence patterns across time periods, disciplines, journals, or any categorical variable. Each group gets its own similarity computation, meaning item frequencies are group-specific. All other parameters (`similarity`, `scale`, `threshold`, `min_occur`, `top_n`) apply per group.

```r
papers <- data.frame(
  year = c(2020, 2020, 2020, 2021, 2021, 2021),
  keywords = c("network; graph; matrix", "graph; algebra",
               "network; algebra; graph",
               "deep learning; nlp", "nlp; transformers",
               "deep learning; transformers; nlp")
)

co(papers, field = "keywords", sep = ";", split_by = "year",
   similarity = "jaccard")
#> # cooccurrence: 7 nodes, 8 edges | split_by: year (2 groups) | similarity: jaccard
#>           from           to    weight count group
#>        algebra        graph 0.6666667     2  2020
#>          graph      network 0.6666667     2  2020
#>  deep learning          nlp 0.6666667     2  2021
#>            nlp transformers 0.6666667     2  2021
#>            ...
```

## Output

`cooccurrence()` returns a tidy data frame of class `cooccurrence` that can be piped, filtered, and joined like any standard data frame. The raw co-occurrence count is always preserved in the `count` column regardless of similarity or scaling, meaning the original counts can always be traced back.

The full matrix, item frequencies, and all parameters are stored as attributes on the returned data frame, making it easy to access the underlying data for further analysis or inspection.

```r
attr(result, "matrix")          # Normalized weight matrix
attr(result, "raw_matrix")      # Raw count matrix (diagonal zeroed)
attr(result, "items")           # Character vector of all items
attr(result, "frequencies")     # Named vector of item frequencies
attr(result, "similarity")      # Similarity measure used
attr(result, "scale")           # Scaling method used
attr(result, "n_transactions")  # Number of transactions
attr(result, "n_items")         # Number of unique items
```

The `cooccurrence` object can be also printed, summarized, and plotted directly as a co-occurrence network (Saqr et al., 2023):

```r
# Summary statistics
summary(result)

# Heatmap
plot(result)

# Network plot (requires igraph)
plot(result, type = "network")
```

### Output formats

The `output` argument controls the format returned directly:

- `"default"` — tidy data frame with `from`, `to`, `weight`, `count` columns
- `"gephi"` — Gephi-ready format with `Source`, `Target`, `Weight`, `Type`, `Count` columns, writable directly to CSV
- `"igraph"` — igraph object for network analysis and visualization
- `"cograph"` — cograph object for use with the cograph package
- `"matrix"` — square co-occurrence matrix

```r
# Default: tidy data frame with from, to, weight, count
co(data, field = "keywords", sep = ";")

# Gephi-ready: Source, Target, Weight, Type, Count columns
co(data, field = "keywords", sep = ";", output = "gephi")
#>   Source  Target Weight       Type Count
#>    graph network      3 Undirected     3

# igraph object
g <- co(data, field = "keywords", sep = ";", output = "igraph")

# cograph object
net <- co(data, field = "keywords", sep = ";", output = "cograph")

# Square matrix
mat <- co(data, field = "keywords", sep = ";", output = "matrix")
```

The Gephi output can be written directly to CSV for import:

```r
write.csv(co(data, field = "keywords", sep = ";", output = "gephi"),
          "network.csv", row.names = FALSE)
```

## Converters

A `cooccurrence` result can be converted to other network formats using the built-in converter functions. All converter packages are optional — install only what you need.

### Matrix

```r
# Normalized similarity matrix
as_matrix(result)

# Raw co-occurrence count matrix
as_matrix(result, type = "raw")
```

### igraph

```r
# install.packages("igraph")
g <- as_igraph(result)
plot(g, edge.width = igraph::E(g)$weight * 3)
igraph::degree(g)
igraph::betweenness(g)
```

### tidygraph

```r
# install.packages("tidygraph")
tg <- as_tidygraph(result)
# Use with ggraph
```

### cograph

```r
# remotes::install_github("mohsaqr/cograph")
net <- as_cograph(result)
cograph::splot(net)
cograph::communities(net)
```

### Nestimate

```r
# remotes::install_github("mohsaqr/Nestimate")
net <- as_netobject(result)
Nestimate::centrality(net)
Nestimate::bootstrap_network(net)
```

## How it works

Regardless of input format, the internal pipeline is:

1. **Parse** input into a list of character vectors (transactions)
2. **Filter** entities below `min_occur` frequency
3. **Build** a binary transaction matrix $B$ (rows = transactions, columns = items)
4. **Compute** raw co-occurrence: $C = B^\top B$ via `crossprod()`
5. **Normalize** using the chosen `similarity` measure
6. **Scale** weights if `scale` is specified
7. **Filter** edges below `threshold` and keep `top_n`
8. **Return** upper triangle as a tidy sorted edge data frame

The computation is vectorized throughout - no loops in the hot path. `crossprod()` delegates to optimized BLAS routines for the matrix multiplication.

## Full parameter reference

| Argument | Type | Description | Default |
|----------|------|-------------|---------|
| `data` | various | Input data (data.frame, matrix, or list) | required |
| `field` | character | Column(s) containing entities (nodes) | `NULL` |
| `by` | character | Column grouping entities into transactions | `NULL` |
| `weight_by` | character | Column with numeric weights for long format (e.g. LDA topic probabilities) | `NULL` |
| `sep` | character | Delimiter for splitting delimited fields | `NULL` |
| `split_by` | character | Column to split data by (separate network per group) | `NULL` |
| `similarity` | character | Normalization measure | `"none"` |
| `counting` | character | `"full"` or `"fractional"` | `"full"` |
| `scale` | character | Weight scaling method | `NULL` |
| `threshold` | numeric | Minimum edge weight (after normalization + scaling) | `0` |
| `min_occur` | integer | Minimum entity frequency (transactions) | `1` |
| `top_n` | integer | Keep only the top N edges by weight (per group if split) | `NULL` |
| `output` | character | Output format: `"default"`, `"gephi"`, `"igraph"`, `"cograph"`, `"matrix"` | `"default"` |

## References

Saqr, M., López-Pernas, S., Conde, M. Á., & Hernández-García, Á. (2023). Social Network Analysis: A Primer, a Guide and a Tutorial in R. In Saqr, M. & López-Pernas, S. (Eds.), *Learning Analytics Methods and Tutorials: A Practical Guide Using R*. Springer. <https://lamethods.org/book1/chapters/ch15-sna/ch15-sna.html>

van Eck, N. J., & Waltman, L. (2009). How to normalize cooccurrence data? An analysis of some well‐known similarity measures. *Journal of the American Society for Information Science and Technology*, *60*(8), 1635-1651.

## Authors

**Mohammed Saqr** — University of Eastern Finland · [saqr.me](https://saqr.me)

**Sonsoles López-Pernas** — University of Eastern Finland · [sonsoles.me](https://sonsoles.me)

**Kamila Misiejuk** — FernUniversität in Hagen · [kamilamisiejuk.com](https://kamilamisiejuk.com)

## License

MIT

library(shiny)
library(cooccure)
library(DT)

# Allow uploads up to 100 MB (default is 5 MB).
# Note: if served behind nginx, also set `client_max_body_size 100m;`
# in the server block — otherwise nginx returns 413 before Shiny sees the request.
options(shiny.maxRequestSize = 100 * 1024^2)

data("movies",       package = "cooccure", envir = environment())
data("actors",       package = "cooccure", envir = environment())
data("actor_genres", package = "cooccure", envir = environment())
data("demo",         package = "cooccure", envir = environment())

# ---- helper: build filtered cograph object ----
# Returns a list(status, value, message). status is one of:
#   "ok"      — value is the cograph_network
#   "empty"   — no rows survived the weight filter
#   "error"   — as_cograph() failed; message has the reason
.filtered_cograph <- function(result, min_w) {
  r <- result[result$weight >= min_w, ]
  if (nrow(r) == 0L)
    return(list(status = "empty", value = NULL, message = NULL))
  # Null stored matrices so as_cograph rebuilds from the filtered edge rows
  attr(r, "matrix")     <- NULL
  attr(r, "raw_matrix") <- NULL
  tryCatch(
    list(status = "ok", value = as_cograph(r), message = NULL),
    error = function(e) list(status = "error", value = NULL,
                             message = conditionMessage(e))
  )
}


# ---- UI ----

ui <- fluidPage(
  tags$head(tags$style(HTML("
    body { font-size: 14px; }

    /* ── Title bar ── */
    .app-title {
      padding: 16px 24px 12px;
      background: linear-gradient(135deg, #1a56db 0%, #2c7be5 60%, #4facfe 100%);
      margin-bottom: 0;
    }
    .app-title-main {
      display: block;
      font-size: 28px; font-weight: 700; letter-spacing: -0.5px;
      color: #fff; font-family: 'Segoe UI', Helvetica, Arial, sans-serif;
    }
    .app-title-sub {
      display: block;
      font-size: 13px; font-weight: 400; letter-spacing: 0.5px;
      color: rgba(255,255,255,0.80); margin-top: 2px;
      font-family: 'Segoe UI', Helvetica, Arial, sans-serif;
    }

    /* ── Sidebar ── */
    .sidebar-panel { background: #f8f9fa; padding: 15px; border-radius: 6px; }
    .section-header { font-weight: 600; margin-top: 14px; margin-bottom: 4px;
                      color: #333; border-bottom: 1px solid #dee2e6; padding-bottom: 3px; }
    .btn-primary  { background-color: #2c7be5; border-color: #2c7be5; }
    .btn-run      { margin-top: 6px; margin-bottom: 10px; }
    .export-strip { margin-top: 18px; padding: 12px; background: #f0f4fb;
                    border-radius: 6px; border: 1px solid #d0dff5; }
    .export-strip h5 { margin-top: 0; margin-bottom: 10px; color: #2c7be5; }

    /* ── Help page ── */
    .help-section {
      background: #fff; border: 1px solid #e4e8ee; border-radius: 8px;
      padding: 20px 24px; margin-bottom: 18px;
    }
    .help-section h4 {
      color: #1a56db; font-weight: 700; margin-top: 0; margin-bottom: 10px;
      border-bottom: 2px solid #e8eff9; padding-bottom: 6px;
    }
    .help-section p, .help-section li { color: #444; line-height: 1.65; }
    .help-section code {
      background: #f0f4fb; color: #c7254e;
      padding: 1px 5px; border-radius: 3px; font-size: 92%;
    }
    .measure-table { width: 100%; border-collapse: collapse; font-size: 13px; }
    .measure-table th {
      background: #1a56db; color: #fff;
      padding: 7px 12px; text-align: left; font-weight: 600;
    }
    .measure-table td { padding: 6px 12px; border-bottom: 1px solid #eee; }
    .measure-table tr:last-child td { border-bottom: none; }
    .measure-table tr:nth-child(even) td { background: #f8f9ff; }
    .badge-fmt {
      display: inline-block; padding: 2px 8px; border-radius: 12px;
      font-size: 11px; font-weight: 600; margin-right: 4px;
      background: #e8eff9; color: #1a56db;
    }

    /* ── Footer ── */
    html, body { height: 100%; }
    body { display: flex; flex-direction: column; min-height: 100vh; }
    .container-fluid { flex: 1; }
    .app-footer {
      padding: 18px 24px 14px;
      margin-top: 30px;
      border-top: 2px solid #e4e8ee;
      text-align: center;
      background: #fafbfd;
    }
    .app-footer .footer-authors {
      font-size: 15px; font-weight: 600; color: #2c3e50;
      margin-bottom: 6px;
    }
    .app-footer .footer-authors a {
      color: #1a56db; text-decoration: none;
    }
    .app-footer .footer-authors a:hover { text-decoration: underline; }
    .app-footer .footer-authors .sep { margin: 0 10px; color: #adb5bd; font-weight: 400; }
    .app-footer .footer-refs {
      font-size: 11px; color: #adb5bd; line-height: 2;
    }
    .app-footer .footer-refs a {
      color: #adb5bd; text-decoration: none; border-bottom: 1px dotted #ced4da;
    }
    .app-footer .footer-refs a:hover { color: #495057; border-bottom-color: #495057; }
    .app-footer .footer-refs .sep { margin: 0 8px; color: #dee2e6; }
  "))),

  tags$div(class = "app-title",
    tags$span(class = "app-title-main", "cooccure"),
    tags$span(class = "app-title-sub", "Co-occurrence Network Explorer")
  ),

  sidebarLayout(
    sidebarPanel(
      width = 3,
      class = "sidebar-panel",

      # ---- Data ----
      div(class = "section-header", "Data"),
      radioButtons("data_source", label = NULL,
                   choices = c("Upload CSV"              = "upload",
                               "Built-in: demo"          = "demo",
                               "Built-in: movies"        = "movies",
                               "Built-in: actors"        = "actors",
                               "Built-in: actor genres"  = "actor_genres"),
                   selected = "upload"),

      conditionalPanel(
        condition = "input.data_source == 'upload'",
        fileInput("file", NULL, accept = c(".csv", "text/csv"),
                  placeholder = "Choose CSV file")
      ),

      # ---- Columns ----
      div(class = "section-header", "Columns"),
      uiOutput("ui_field"),
      uiOutput("ui_by"),
      uiOutput("ui_split_by"),
      uiOutput("ui_sep"),

      # ---- BUILD NETWORK (early) ----
      actionButton("run", "Build network", class = "btn-primary btn-run",
                   width = "100%", icon = icon("play")),

      # ---- Similarity & Counting ----
      div(class = "section-header", "Similarity"),
      selectInput("similarity", label = NULL,
                  choices = c("none", "jaccard", "cosine", "inclusion",
                              "association", "dice", "equivalence", "relative"),
                  selected = "none"),

      div(class = "section-header", "Counting"),
      radioButtons("counting", label = NULL,
                   choices = c("full", "fractional"), selected = "full",
                   inline = TRUE),

      # ---- Filters ----
      div(class = "section-header", "Filters"),
      fluidRow(
        column(6, numericInput("min_occur", "Min freq",  value = 1, min = 1, step = 1)),
        column(6, numericInput("threshold", "Threshold", value = 0, min = 0, step = 0.01))
      ),
      numericInput("top_n", "Top N edges (0 = all)", value = 0, min = 0, step = 25),

      # ---- Scale ----
      div(class = "section-header", "Scale (optional)"),
      selectInput("scale", label = NULL,
                  choices = c("none", "minmax", "log", "log10",
                              "binary", "zscore", "sqrt", "proportion"),
                  selected = "none")
    ),

    mainPanel(
      width = 9,
      tabsetPanel(
        id = "tabs",

        # ---- Quick Start (landing) ----
        tabPanel("Quick Start", value = "quickstart",
          br(),
          fluidRow(
            column(6, offset = 3,
              div(class = "help-section",
                tags$h4("\u26a1 Get started in 3 steps"),
                tags$ol(style = "font-size: 15px; line-height: 2;",
                  tags$li("Pick a dataset in the sidebar — upload your own CSV or try a built-in demo (", tags$em("movies"), " or ", tags$em("actors"), ")."),
                  tags$li("Select the ", tags$strong("Field"), " column that holds your items. Set ", tags$strong("Separator"), " if items are comma-separated inside one cell, or ", tags$strong("Group by"), " if each row is one item in a transaction."),
                  tags$li("Click ", tags$strong("Build network"), ", then explore results in the ", tags$strong("Summary"), ", ", tags$strong("View edge table"), ", and ", tags$strong("Network"), " tabs.")
                )
              ),
              div(class = "help-section",
                tags$h4("\u2139\ufe0f Key options"),
                tags$table(class = "measure-table",
                  tags$thead(tags$tr(tags$th("Option"), tags$th("What it does"))),
                  tags$tbody(
                    tags$tr(tags$td(tags$code("Similarity")), tags$td("How edge weights are calculated. Jaccard is a safe default for most datasets.")),
                    tags$tr(tags$td(tags$code("Min freq")),   tags$td("Drop rare items that appear in fewer than N transactions.")),
                    tags$tr(tags$td(tags$code("Threshold")),  tags$td("Hide weak edges below this weight.")),
                    tags$tr(tags$td(tags$code("Top N edges")), tags$td("Keep only the N strongest edges — useful for large networks.")),
                    tags$tr(tags$td(tags$code("split_by")),   tags$td("Build one network per group (e.g. per year) in a single run."))
                  )
                )
              ),
              tags$p(style = "text-align:center; color:#adb5bd; font-size:12px; margin-top:8px;",
                "For input format details and similarity measure definitions, see the ",
                tags$strong("Help"), " tab."
              )
            )
          )
        ),

        # ---- Help (reference) ----
        tabPanel("Help", value = "help",
          br(),
          fluidRow(
            column(6,
              div(class = "help-section",
                tags$h4("\U0001f4c2 Input formats"),
                tags$p("The app auto-detects format from your data. Guide it with ", tags$code("sep"), " and ", tags$code("by"), " if needed."),
                tags$table(class = "measure-table",
                  tags$thead(tags$tr(tags$th("Format"), tags$th("Description"), tags$th("Example"))),
                  tags$tbody(
                    tags$tr(tags$td(tags$span(class = "badge-fmt", "delimited")),
                            tags$td("Multiple items in one cell, separated by a delimiter"),
                            tags$td(tags$code("\"Drama,Action\""))),
                    tags$tr(tags$td(tags$span(class = "badge-fmt", "long")),
                            tags$td("One item per row; a second column groups them"),
                            tags$td(tags$code("actor + movie_id"))),
                    tags$tr(tags$td(tags$span(class = "badge-fmt", "wide")),
                            tags$td("Each row is a transaction; columns are items"),
                            tags$td("survey response columns")),
                    tags$tr(tags$td(tags$span(class = "badge-fmt", "binary")),
                            tags$td("Wide 0/1 presence matrix"),
                            tags$td("term-document matrix")),
                    tags$tr(tags$td(tags$span(class = "badge-fmt", "multi_delimited")),
                            tags$td("Multiple delimited columns merged before counting"),
                            tags$td(tags$code("kw_1, kw_2")))
                  )
                )
              )
            ),
            column(6,
              div(class = "help-section",
                tags$h4("\U0001f4cf Similarity measures"),
                tags$table(class = "measure-table",
                  tags$thead(tags$tr(tags$th("Measure"), tags$th("Best for"))),
                  tags$tbody(
                    tags$tr(tags$td(tags$strong("none")),        tags$td("Raw counts — frequency-based analyses")),
                    tags$tr(tags$td(tags$strong("jaccard")),     tags$td("General-purpose; robust to unequal frequencies")),
                    tags$tr(tags$td(tags$strong("cosine")),      tags$td("Text and keyword networks")),
                    tags$tr(tags$td(tags$strong("inclusion")),   tags$td("Subset / hierarchical relationships")),
                    tags$tr(tags$td(tags$strong("association")), tags$td("Market-basket / association rules")),
                    tags$tr(tags$td(tags$strong("dice")),        tags$td("Balanced co-occurrence (harmonic Jaccard)")),
                    tags$tr(tags$td(tags$strong("equivalence")), tags$td("Bibliographic coupling / citation networks")),
                    tags$tr(tags$td(tags$strong("relative")),    tags$td("Comparing across different-sized corpora"))
                  )
                )
              ),
              div(class = "help-section",
                tags$h4("\U0001f4be Export formats"),
                tags$ul(
                  tags$li(tags$strong("CSV"), " — standard edge list for R / Python."),
                  tags$li(tags$strong("CSV (Gephi)"), " — ", tags$code("Source / Target / Weight"), " columns for direct Gephi import."),
                  tags$li(tags$strong("GraphML"), " — interoperable XML format for Gephi, Cytoscape, NetworkX.")
                )
              )
            )
          )
        ),

        # ---- Summary ----
        tabPanel("Summary",
          br(),
          verbatimTextOutput("summary_out"),
          br(),
          verbatimTextOutput("print_out"),

          # Early export strip
          uiOutput("export_strip")
        ),

        # ---- View edge table ----
        tabPanel("View edge table",
          br(),
          DTOutput("edge_table")
        ),

        # ---- Network (cograph) ----
        tabPanel("Network",
          br(),
          fluidRow(
            column(3,
              sliderInput("min_edge_w", "Min edge weight",
                          min = 0, max = 1, value = 0, step = 0.01,
                          width = "100%")
            ),
            column(3,
              selectInput("net_layout", "Layout",
                          choices = c("Gephi"                = "gephi",
                                      "Fruchterman-Reingold" = "fr",
                                      "Kamada-Kawai"         = "kk",
                                      "Circle"               = "circle",
                                      "Nicely"               = "nicely"),
                          selected = "gephi", width = "100%")
            ),
            column(3,
              numericInput("label_size", "Label size", value = 0.8,
                           min = 0.3, max = 2, step = 0.1, width = "100%")
            ),
            column(3,
              numericInput("edge_width_max", "Max edge width", value = 4,
                           min = 0.5, max = 10, step = 0.5, width = "100%")
            )
          ),
          uiOutput("net_ui")
        ),

        # ---- Export ----
        tabPanel("Export",
          br(),
          h5("Edge list"),
          downloadButton("dl_csv2",     "CSV (default)"),
          " ",
          downloadButton("dl_gephi2",   "CSV (Gephi)"),
          br(), br(),
          h5("Network file"),
          downloadButton("dl_graphml2", "GraphML (igraph)")
        )
      )
    )
  ),

  tags$footer(class = "app-footer",
    tags$div(class = "footer-authors",
      tags$a(href = "https://saqr.me",     target = "_blank", "Mohammed Saqr"),
      tags$span(class = "sep", "\u00b7"),
      tags$a(href = "https://sonsoles.me", target = "_blank", "Sonsoles L\u00f3pez-Pernas"),
      tags$span(class = "sep", "\u00b7"),
      tags$a(href = "https://kamilamisiejuk.com", target = "_blank", "Kamila Misiejuk")
    ),
    tags$div(class = "footer-refs",
      tags$a(href = "https://lamethods.org/book1/chapters/ch15-sna/ch15-sna.html",
             target = "_blank",
             "Social Network Analysis: A Primer, a Guide and a Tutorial in R"),
      tags$span(class = "sep", "\u00b7"),
      tags$a(href = "https://link.springer.com/chapter/10.1007/978-3-031-25336-2_5",
             target = "_blank",
             "Scientometrics: A Concise Introduction and a Detailed Methodology")
    )
  )
)


# ---- Server ----

server <- function(input, output, session) {

  # ---- reactive: loaded data ----
  data_loaded <- reactive({
    switch(input$data_source,
      movies       = movies,
      actors       = actors,
      actor_genres = actor_genres,
      demo         = demo,
      upload = {
        req(input$file)
        read.csv(input$file$datapath, stringsAsFactors = FALSE)
      }
    )
  })

  # ---- pre-fill columns when source changes ----
  observeEvent(data_loaded(), {
    d        <- data_loaded()
    cols     <- colnames(d)
    cols_opt <- c("— none —" = "", cols)

    default_field <- switch(input$data_source,
      movies       = "genres",
      actors       = "actor",
      actor_genres = "actor",
      demo         = "actor",
      cols[1]
    )
    updateSelectInput(session, "field_sel", choices = cols, selected = default_field)

    default_sep <- if (input$data_source == "movies") "," else ""
    updateTextInput(session, "sep_val", value = default_sep)

    default_by <- switch(input$data_source,
      actors       = "tconst",
      actor_genres = "genre",
      demo         = "movie",
      ""
    )
    updateSelectInput(session, "by_sel",       choices = cols_opt, selected = default_by)
    updateSelectInput(session, "split_by_sel", choices = cols_opt, selected = "")
    updateNumericInput(session, "min_occur", value = 1L)
  })

  # ---- dynamic column selectors ----
  output$ui_field <- renderUI({
    d <- data_loaded()
    selectInput("field_sel", "Field (nodes)", choices = colnames(d))
  })

  output$ui_sep <- renderUI({
    textInput("sep_val", "Separator (sep)", value = "",
              placeholder = "e.g.  ,  ;  |")
  })

  output$ui_by <- renderUI({
    d <- data_loaded()
    selectInput("by_sel", "Group by (by)",
                choices = c("— none —" = "", colnames(d)), selected = "")
  })

  output$ui_split_by <- renderUI({
    d <- data_loaded()
    selectInput("split_by_sel", "Split by",
                choices = c("— none —" = "", colnames(d)), selected = "")
  })

  # ---- reactive: cooccurrence result ----
  result <- eventReactive(input$run, {
    d        <- data_loaded()
    field    <- input$field_sel
    sep      <- if (nzchar(trimws(input$sep_val))) trimws(input$sep_val) else NULL
    by       <- if (nzchar(input$by_sel))       input$by_sel       else NULL
    split_by <- if (nzchar(input$split_by_sel)) input$split_by_sel else NULL
    top_n    <- if (input$top_n > 0) as.integer(input$top_n) else NULL
    scale    <- if (input$scale == "none") NULL else input$scale

    withProgress(message = "Computing co-occurrences…", value = 0.5, {
      tryCatch(
        cooccurrence(d,
                     field      = field,
                     sep        = sep,
                     by         = by,
                     split_by   = split_by,
                     similarity = input$similarity,
                     counting   = input$counting,
                     scale      = scale,
                     threshold  = input$threshold,
                     min_occur  = as.integer(input$min_occur),
                     top_n      = top_n),
        error = function(e) {
          showNotification(conditionMessage(e), type = "error", duration = 8)
          NULL
        }
      )
    })
  })

  # update slider range to match result weights
  observeEvent(result(), {
    r <- result()
    if (!is.null(r) && nrow(r) > 0) {
      mn_raw <- min(r$weight, na.rm = TRUE)
      mx_raw <- max(r$weight, na.rm = TRUE)
      # Floor the slider min below the true minimum so the initial
      # filter (value >= min) includes every edge. Rounding up would
      # silently drop edges at the minimum weight.
      mn <- floor(mn_raw * 10000) / 10000
      mx <- ceiling(mx_raw * 10000) / 10000
      step <- if (mx > mn) round((mx - mn) / 100, 4) else 0.01
      updateSliderInput(session, "min_edge_w",
                        min   = mn, max = mx, value = mn,
                        step  = step)
    }
    # Switch to Summary tab so results are immediately visible
    updateTabsetPanel(session, "tabs", selected = "Summary")
  })

  # ---- Summary tab ----
  output$summary_out <- renderPrint({
    req(result())
    summary(result())
  })

  output$print_out <- renderPrint({
    req(result())
    print(result(), n = 10L)
  })

  output$export_strip <- renderUI({
    req(result())
    div(class = "export-strip",
      h5("Export"),
      downloadButton("dl_csv",     "CSV"),
      " ",
      downloadButton("dl_gephi",   "Gephi CSV"),
      " ",
      downloadButton("dl_graphml", "GraphML")
    )
  })

  # ---- View edge table tab ----
  output$edge_table <- renderDT({
    req(result())
    r <- as.data.frame(result())
    r$weight <- round(r$weight, 6)
    datatable(r,
              filter   = "top",
              rownames = FALSE,
              options  = list(pageLength = 25, scrollX = TRUE))
  })

  # ---- Network tab (cograph) ----
  output$net_ui <- renderUI({
    req(result())
    r          <- result()
    min_w      <- input$min_edge_w
    filtered_n <- sum(r$weight >= min_w, na.rm = TRUE)
    has_cograph <- requireNamespace("cograph", quietly = TRUE)

    diag <- tags$div(
      style = "font-family: monospace; font-size: 12px; color: #666;
               background: #f8f9fa; padding: 8px 12px; margin-bottom: 12px;
               border-left: 3px solid #2c7be5;",
      sprintf(
        "diagnostics  edges=%d  min_weight=%.6f  max_weight=%.6f  filter_threshold=%.6f  passing=%d  cograph_installed=%s",
        nrow(r), min(r$weight, na.rm = TRUE), max(r$weight, na.rm = TRUE),
        as.numeric(min_w), filtered_n, has_cograph
      )
    )

    if (!has_cograph) {
      return(tagList(diag, tags$div(class = "alert alert-danger",
        tags$strong("Package 'cograph' is not installed on this server."),
        tags$br(),
        "Install it with ",
        tags$code("install.packages('cograph', repos = 'https://mohsaqr.r-universe.dev')"),
        " on the host, then restart the app.")))
    }
    res <- .filtered_cograph(r, min_w)

    body <- if (res$status == "empty") {
      tags$div(class = "alert alert-warning",
               sprintf("No edges with weight >= %.4f. Lower the minimum weight filter.",
                       as.numeric(min_w)))
    } else if (res$status == "error") {
      tags$div(class = "alert alert-danger",
               tags$strong("Could not build the network: "),
               res$message)
    } else if (res$value$n_edges > 3000) {
      tags$div(class = "alert alert-info",
               sprintf("%d edges — too many to render clearly. Raise the minimum weight filter.",
                       res$value$n_edges))
    } else {
      plotOutput("cograph_plot", height = "600px")
    }
    tagList(diag, body)
  })

  output$cograph_plot <- renderPlot({
    req(result())
    req(requireNamespace("cograph", quietly = TRUE))
    res <- .filtered_cograph(result(), input$min_edge_w)
    req(res$status == "ok")

    cograph::splot(
      res$value,
      layout          = input$net_layout,
      scale_nodes_by  = "degree",
      label_size      = input$label_size,
      edge_width_range = c(0.1, input$edge_width_max)
    )
  })

  # ---- Download handlers (shared by Summary strip + Export tab) ----
  output$dl_csv <- downloadHandler(
    filename = function() paste0("cooccurrence_", Sys.Date(), ".csv"),
    content  = function(f) write.csv(as.data.frame(result()), f, row.names = FALSE)
  )

  output$dl_gephi <- downloadHandler(
    filename = function() paste0("cooccurrence_gephi_", Sys.Date(), ".csv"),
    content  = function(f) {
      req(result())
      gephi <- tryCatch(
        cooccurrence(data_loaded(),
                     field      = input$field_sel,
                     sep        = if (nzchar(trimws(input$sep_val))) trimws(input$sep_val) else NULL,
                     by         = if (nzchar(input$by_sel))       input$by_sel       else NULL,
                     split_by   = if (nzchar(input$split_by_sel)) input$split_by_sel else NULL,
                     similarity = input$similarity,
                     counting   = input$counting,
                     scale      = if (input$scale == "none") NULL else input$scale,
                     threshold  = input$threshold,
                     min_occur  = as.integer(input$min_occur),
                     top_n      = if (input$top_n > 0) as.integer(input$top_n) else NULL,
                     output     = "gephi"),
        error = function(e) NULL
      )
      write.csv(as.data.frame(gephi), f, row.names = FALSE)
    }
  )

  output$dl_graphml <- downloadHandler(
    filename = function() paste0("cooccurrence_", Sys.Date(), ".graphml"),
    content  = function(f) {
      req(result())
      if (!requireNamespace("igraph", quietly = TRUE)) {
        showNotification("igraph is required for GraphML export.", type = "error")
        return(NULL)
      }
      igraph::write_graph(as_igraph(result()), f, format = "graphml")
    }
  )

  # ---- Export tab handlers (separate IDs to avoid duplicate-ID conflict) ----
  output$dl_csv2 <- downloadHandler(
    filename = function() paste0("cooccurrence_", Sys.Date(), ".csv"),
    content  = function(f) write.csv(as.data.frame(result()), f, row.names = FALSE)
  )

  output$dl_gephi2 <- downloadHandler(
    filename = function() paste0("cooccurrence_gephi_", Sys.Date(), ".csv"),
    content  = function(f) {
      req(result())
      gephi <- tryCatch(
        cooccurrence(data_loaded(),
                     field      = input$field_sel,
                     sep        = if (nzchar(trimws(input$sep_val))) trimws(input$sep_val) else NULL,
                     by         = if (nzchar(input$by_sel))       input$by_sel       else NULL,
                     split_by   = if (nzchar(input$split_by_sel)) input$split_by_sel else NULL,
                     similarity = input$similarity,
                     counting   = input$counting,
                     scale      = if (input$scale == "none") NULL else input$scale,
                     threshold  = input$threshold,
                     min_occur  = as.integer(input$min_occur),
                     top_n      = if (input$top_n > 0) as.integer(input$top_n) else NULL,
                     output     = "gephi"),
        error = function(e) NULL
      )
      write.csv(as.data.frame(gephi), f, row.names = FALSE)
    }
  )

  output$dl_graphml2 <- downloadHandler(
    filename = function() paste0("cooccurrence_", Sys.Date(), ".graphml"),
    content  = function(f) {
      req(result())
      if (!requireNamespace("igraph", quietly = TRUE)) {
        showNotification("igraph is required for GraphML export.", type = "error")
        return(NULL)
      }
      igraph::write_graph(as_igraph(result()), f, format = "graphml")
    }
  )
}

shinyApp(ui, server)

# ---- Co-occurrence Network Construction ----

# TraMineR/tna void markers
.void_markers <- c("%", "*", "", "NA", "NaN")

#' Build a co-occurrence network
#'
#' Constructs an undirected co-occurrence network from various input formats
#' and returns a tidy edge data frame. Argument names follow the citenets
#' convention.
#'
#' @param data Input data. Accepts:
#'   \itemize{
#'     \item A \code{data.frame} with a delimited column (\code{field} + \code{sep}).
#'     \item A \code{data.frame} in long/bipartite format (\code{field} + \code{by}).
#'     \item A binary (0/1) \code{data.frame} or \code{matrix} (auto-detected).
#'     \item A wide sequence \code{data.frame} or \code{matrix} (non-binary).
#'     \item A \code{list} of character vectors (each element is a transaction).
#'   }
#' @param field Character. The entity column --- determines what the nodes are.
#'   For delimited format, a single column split by \code{sep}. For
#'   long/bipartite, the item column. For multi-column delimited, a vector
#'   of column names pooled per row. Use \code{field = "all"} for wide
#'   sequence data (e.g. TraMineR / tna format) where every column is a
#'   time point and cell values are the items.
#' @param by Character or \code{NULL}. Grouping column for long/bipartite
#'   format. Each unique value defines one transaction.
#' @param weight_by Character or \code{NULL}. Column name containing a numeric
#'   association strength for each entity-transaction pair. Only accepted for
#'   long format (\code{field} + \code{by}). When supplied, each entity
#'   contributes its weight rather than 1, so
#'   \eqn{C_{ij} = \sum_d w_{id} \cdot w_{jd}}.
#'   Typical use: topic-document probability matrices from LDA or similar
#'   models.
#' @param sep Character or \code{NULL}. Separator for splitting delimited
#'   fields.
#' @param split_by Character or \code{NULL}. Column name to split the data
#'   by before computing co-occurrence. A separate network is computed per
#'   group and the results are combined into a single data frame with an
#'   additional \code{group} column. Only works with data.frame inputs.
#' @param similarity Character. Similarity measure:
#'   \describe{
#'     \item{\code{"none"}}{Raw co-occurrence counts.}
#'     \item{\code{"jaccard"}}{\eqn{C_{ij} / (f_i + f_j - C_{ij})}.}
#'     \item{\code{"cosine"}}{Salton's cosine:
#'       \eqn{C_{ij} / \sqrt{f_i \cdot f_j}}.}
#'     \item{\code{"inclusion"}}{Simpson coefficient:
#'       \eqn{C_{ij} / \min(f_i, f_j)}.}
#'     \item{\code{"association"}}{Association strength:
#'       \eqn{C_{ij} / (f_i \cdot f_j)}
#'       (van Eck & Waltman, 2009).}
#'     \item{\code{"dice"}}{\eqn{2 C_{ij} / (f_i + f_j)}.}
#'     \item{\code{"equivalence"}}{Salton's cosine squared:
#'       \eqn{C_{ij}^2 / (f_i \cdot f_j)}.}
#'     \item{\code{"relative"}}{Row-normalized: each row sums to 1.}
#'   }
#' @param scale Character or \code{NULL}. Optional scaling applied to weights
#'   after similarity normalization:
#'   \describe{
#'     \item{\code{NULL} or \code{"none"}}{No scaling.}
#'     \item{\code{"minmax"}}{Min-max to \eqn{[0, 1]}.}
#'     \item{\code{"log"}}{Natural log: \eqn{\log(1 + w)}.}
#'     \item{\code{"log10"}}{Log base 10: \eqn{\log_{10}(1 + w)}.}
#'     \item{\code{"binary"}}{Binary: 1 if \eqn{w > 0}, else 0.}
#'     \item{\code{"zscore"}}{Z-score standardization.}
#'     \item{\code{"sqrt"}}{Square root.}
#'     \item{\code{"proportion"}}{Divide by sum of all weights.}
#'   }
#' @param counting Character. Counting method:
#'   \describe{
#'     \item{\code{"full"}}{Each co-occurring pair adds 1 regardless of
#'       transaction size. Default.}
#'     \item{\code{"fractional"}}{Each pair adds \eqn{1 / (n_i - 1)}
#'       where \eqn{n_i} is the number of items in transaction \eqn{i}.
#'       Transactions with many items contribute less per pair
#'       (Perianes-Rodriguez et al., 2016).}
#'   }
#' @param threshold Numeric. Minimum edge weight to retain. Applied after
#'   similarity and scaling. Default 0.
#' @param min_occur Integer. Minimum entity frequency. Entities appearing in
#'   fewer than \code{min_occur} transactions are dropped. Default 1.
#' @param top_n Integer or \code{NULL}. Keep only the top \code{top_n} edges
#'   by weight. When \code{split_by} is used, applied per group.
#'   Default \code{NULL} (all edges).
#' @param output Character. Column naming convention for the output:
#'   \describe{
#'     \item{\code{"default"}}{\code{from}, \code{to}, \code{weight}, \code{count}.}
#'     \item{\code{"gephi"}}{\code{Source}, \code{Target}, \code{Weight},
#'       \code{Type} (= \code{"Undirected"}). Ready for Gephi import.}
#'     \item{\code{"igraph"}}{Returns an \code{igraph} graph object directly.}
#'     \item{\code{"cograph"}}{Returns a \code{cograph_network} object directly.}
#'     \item{\code{"matrix"}}{Returns the square co-occurrence matrix.}
#'   }
#' @param ... Currently unused.
#'
#' @return Depends on \code{output}:
#'   \itemize{
#'     \item \code{"default"}: A \code{cooccurrence} data frame with columns
#'       \code{from}, \code{to}, \code{weight}, \code{count} (and \code{group}
#'       when \code{split_by} is used).
#'     \item \code{"gephi"}: A data frame with columns \code{Source},
#'       \code{Target}, \code{Weight}, \code{Type}, \code{Count}. Ready for
#'       Gephi CSV import.
#'     \item \code{"igraph"}: An \code{igraph} graph object.
#'     \item \code{"cograph"}: A \code{cograph_network} object.
#'     \item \code{"matrix"}: A square numeric co-occurrence matrix.
#'   }
#'   For the data frame outputs, rows are sorted by weight descending and
#'   attributes store the full matrix, item frequencies, and parameters.
#'
#' @references
#' van Eck, N. J., & Waltman, L. (2009). How to normalize co-occurrence
#' data? An analysis of some well-known similarity measures. \emph{Journal of
#' the American Society for Information Science and Technology}, 60(8),
#' 1635--1651.
#'
#' @examples
#' # Delimited keywords
#' df <- data.frame(
#'   id = 1:4,
#'   keywords = c("network; graph", "graph; matrix; network",
#'                "matrix; algebra", "network; algebra; graph")
#' )
#' cooccurrence(df, field = "keywords", sep = ";")
#'
#' # Split by a grouping variable
#' df$year <- c(2020, 2020, 2021, 2021)
#' cooccurrence(df, field = "keywords", sep = ";", split_by = "year")
#'
#' # List of transactions with Jaccard similarity
#' cooccurrence(list(c("A","B","C"), c("B","C"), c("A","C")),
#'              similarity = "jaccard")
#'
#' # Short alias
#' co(df, field = "keywords", sep = ";", similarity = "cosine")
#'
#' # Weighted long format (e.g. LDA topic-document probabilities)
#' theta <- data.frame(
#'   doc   = c("d1","d1","d1","d2","d2","d3","d3"),
#'   topic = c("T1","T2","T3","T1","T3","T2","T3"),
#'   prob  = c(0.6, 0.3, 0.1, 0.4, 0.6, 0.5, 0.5)
#' )
#' cooccurrence(theta, field = "topic", by = "doc", weight_by = "prob")
#'
#' @export
cooccurrence <- function(data, field = NULL, by = NULL, sep = NULL,
                         weight_by = NULL,
                         split_by = NULL,
                         similarity = c("none", "jaccard", "cosine",
                                        "inclusion", "association",
                                        "dice", "equivalence", "relative"),
                         counting = c("full", "fractional"),
                         scale = NULL,
                         threshold = 0, min_occur = 1L,
                         top_n = NULL,
                         output = c("default", "gephi", "igraph",
                                    "cograph", "matrix"), ...) {
  similarity <- match.arg(similarity)
  counting <- match.arg(counting)
  output <- match.arg(output)
  threshold <- as.numeric(threshold)
  min_occur <- as.integer(min_occur)
  stopifnot(threshold >= 0, min_occur >= 1L)

  if (is.null(scale) || identical(scale, "none")) {
    scale_method <- "none"
  } else {
    scale_method <- match.arg(scale, c("none", "minmax", "log", "log10",
                                       "binary", "zscore", "sqrt",
                                       "proportion"))
  }

  # ---- split_by: compute per group, combine ----
  if (!is.null(split_by)) {
    stopifnot(is.data.frame(data), split_by %in% names(data))
    groups <- split(data, data[[split_by]])
    parts <- lapply(names(groups), function(g) {
      sub <- groups[[g]]
      # Drop the split_by column so it doesn't interfere with format detection
      sub[[split_by]] <- NULL
      edges <- tryCatch(
        .co_core(sub, field = field, by = by, sep = sep,
                 weight_by = weight_by,
                 similarity = similarity, counting = counting,
                 scale_method = scale_method,
                 threshold = threshold, min_occur = min_occur,
                 top_n = top_n),
        error = function(e) NULL
      )
      if (is.null(edges) || nrow(edges) == 0L) return(NULL)
      edges$group <- g
      edges
    })
    parts <- parts[!vapply(parts, is.null, logical(1))]
    if (length(parts) == 0L)
      stop("No groups produced any edges.", call. = FALSE)
    edges <- do.call(rbind, parts)
    rownames(edges) <- NULL

    class(edges) <- c("cooccurrence", "data.frame")
    attr(edges, "similarity") <- similarity
    attr(edges, "scale") <- scale_method
    attr(edges, "threshold") <- threshold
    attr(edges, "min_occur") <- min_occur
    attr(edges, "split_by") <- split_by
    attr(edges, "groups") <- names(groups)
    return(.co_format_output(edges, output))
  }

  # ---- Single-group path ----
  result <- .co_core(data, field = field, by = by, sep = sep,
                     weight_by = weight_by,
                     similarity = similarity, counting = counting,
                     scale_method = scale_method,
                     threshold = threshold, min_occur = min_occur,
                     top_n = top_n)

  .co_format_output(result, output)
}


#' @rdname cooccurrence
#' @export
co <- cooccurrence


# ---- Core pipeline (used by both single and split_by paths) ----

#' @noRd
.co_core <- function(data, field, by, sep, weight_by = NULL, similarity,
                     counting, scale_method, threshold, min_occur, top_n) {
  # Parse input
  fmt <- .co_detect_format(data, field, by, sep)

  # Weighted path — long format only
  if (!is.null(weight_by)) {
    if (fmt != "long")
      stop("`weight_by` is only supported for long format (field + by).",
           call. = FALSE)
    return(.co_core_weighted(data, field, by, weight_by, similarity,
                             scale_method, threshold, min_occur, top_n))
  }

  if (fmt == "field_only")
    .co_warn_missing_sep(data, field)

  transactions <- switch(fmt,
    delimited       = .co_parse_delimited(data, field, sep),
    multi_delimited = .co_parse_multi_delimited(data, field, sep),
    field_only      = .co_parse_field_only(data, field),
    long            = .co_parse_long(data, field, by),
    binary          = .co_parse_binary(data),
    wide            = .co_parse_wide(data),
    list            = .co_parse_list(data)
  )

  # Drop empty transactions
  transactions <- transactions[vapply(transactions, length, integer(1)) > 0L]
  if (length(transactions) == 0L)
    stop("No non-empty transactions found in the input data.", call. = FALSE)

  # min_occur filter
  if (min_occur > 1L) {
    freq_table <- table(unlist(transactions))
    keep <- names(freq_table[freq_table >= min_occur])
    transactions <- lapply(transactions, function(t) t[t %in% keep])
    transactions <- transactions[vapply(transactions, length, integer(1)) > 0L]
    if (length(transactions) == 0L)
      stop("No transactions remain after min_occur filtering.", call. = FALSE)
  }

  # Build sparse bipartite matrix (works x items) with counting weights baked in
  sp <- .co_build_sparse(transactions, counting)
  n_trans <- sp$n
  n_items <- sp$k
  items <- sp$items

  # Counting-weighted co-occurrence (sparse k x k)
  C <- Matrix::crossprod(sp$B)
  # Raw binary co-occurrence (sparse k x k) for the count column and frequencies
  C_raw <- Matrix::crossprod(sp$B_bin)
  freq <- as.numeric(Matrix::colSums(sp$B_bin))
  names(freq) <- items

  # Zero diagonals (self co-occurrence is not an edge)
  Matrix::diag(C) <- 0
  Matrix::diag(C_raw) <- 0
  C <- Matrix::drop0(C)
  C_raw <- Matrix::drop0(C_raw)

  .co_finalize(C, C_raw, freq, n_trans, n_items, items,
               similarity, scale_method, threshold, min_occur, top_n)
}


# ---- Shared finalization: normalize -> scale -> threshold -> edges -> stamp ----

#' Finalize edges from sparse co-occurrence matrices.
#'
#' Operates on triplets throughout — never materialises a dense k x k matrix.
#' For symmetric similarities, the attribute `matrix` is stored as a symmetric
#' sparse `dsCMatrix`; for `similarity = "relative"` it is a general
#' `dgCMatrix` holding both triangles.
#'
#' @param C Sparse counting-weighted co-occurrence matrix (k x k, diag = 0).
#' @param C_raw Sparse raw (binary) co-occurrence matrix (k x k, diag = 0).
#' @param freq Named numeric vector of item frequencies.
#' @param items Character vector of item names (k).
#' @noRd
.co_finalize <- function(C, C_raw, freq, n_trans, n_items, items,
                         similarity, scale_method, threshold, min_occur, top_n) {
  ## Upper-triangle triplets of C (counting-weighted) — these carry x values.
  C_upper_T <- methods::as(Matrix::triu(C, k = 1L), "TsparseMatrix")
  i <- C_upper_T@i + 1L
  j <- C_upper_T@j + 1L
  c_vals <- C_upper_T@x

  if (length(i) == 0L) {
    edges <- data.frame(from = character(0), to = character(0),
                        weight = numeric(0), count = integer(0),
                        stringsAsFactors = FALSE)
    W_sparse <- Matrix::sparseMatrix(
      i = integer(0), j = integer(0), x = numeric(0),
      dims = c(n_items, n_items), symmetric = TRUE,
      dimnames = list(items, items)
    )
  } else {
    ## Raw counts at the same positions. Non-zero pattern of C and C_raw is
    ## identical (both are crossprods of binaries with the same support), so
    ## [cbind(i, j)] indexing returns counts aligned with c_vals.
    raw_vals <- as.integer(C_raw[cbind(i, j)])

    ## Similarity normalisation on triplets.
    if (similarity == "none") {
      W_x <- c_vals
      W_sparse <- Matrix::sparseMatrix(
        i = i, j = j, x = W_x,
        dims = c(n_items, n_items), symmetric = TRUE,
        dimnames = list(items, items)
      )
    } else if (similarity == "relative") {
      ## Asymmetric: W[i,j] = C[i,j] / rowSums(C)[i]. Needs both triangles.
      ## C is a symmetric dsCMatrix whose TsparseMatrix form stores only the
      ## upper triangle, so we mirror (i, j, c_vals) to build the full matrix
      ## explicitly rather than relying on the symmetric packing.
      rs <- as.numeric(Matrix::rowSums(C))
      rs[rs == 0] <- 1
      W_x <- c_vals / rs[i]

      ii <- c(i, j)
      jj <- c(j, i)
      xx <- c(c_vals, c_vals)
      W_full_x <- xx / rs[ii]
    } else {
      denom <- switch(similarity,
        jaccard     = freq[i] + freq[j] - c_vals,
        cosine      = sqrt(freq[i] * freq[j]),
        inclusion   = pmin(freq[i], freq[j]),
        association = freq[i] * freq[j],
        dice        = freq[i] + freq[j],
        equivalence = freq[i] * freq[j]
      )
      denom[denom == 0] <- 1
      numer <- switch(similarity,
        dice        = 2 * c_vals,
        equivalence = c_vals^2,
        c_vals
      )
      W_x <- as.numeric(numer / denom)
      W_sparse <- Matrix::sparseMatrix(
        i = i, j = j, x = W_x,
        dims = c(n_items, n_items), symmetric = TRUE,
        dimnames = list(items, items)
      )
    }

    ## Scaling operates on the full non-zero population of the stored matrix.
    ## For symmetric W, that population is c(W_x, W_x); for relative W, it is
    ## the asymmetric entries in both triangles.
    if (scale_method != "none") {
      if (similarity == "relative") {
        population <- W_full_x
        W_full_x <- .co_scale_values(W_full_x, population, scale_method)
        W_x <- .co_scale_values(W_x, population, scale_method)
      } else {
        population <- c(W_x, W_x)
        W_x <- .co_scale_values(W_x, population, scale_method)
        W_sparse <- Matrix::sparseMatrix(
          i = i, j = j, x = W_x,
          dims = c(n_items, n_items), symmetric = TRUE,
          dimnames = list(items, items)
        )
      }
    }

    ## Build the asymmetric W_sparse for 'relative' (post-scaling).
    if (similarity == "relative") {
      W_sparse <- Matrix::sparseMatrix(
        i = ii, j = jj, x = W_full_x,
        dims = c(n_items, n_items),
        dimnames = list(items, items)
      )
    }

    ## Threshold filter on edge weights (before name lookup is cheap).
    if (threshold > 0) {
      keep <- W_x >= threshold
      W_x <- W_x[keep]; i <- i[keep]; j <- j[keep]; raw_vals <- raw_vals[keep]
    }

    edges <- data.frame(
      from   = items[i],
      to     = items[j],
      weight = W_x,
      count  = raw_vals,
      stringsAsFactors = FALSE
    )
    edges <- edges[order(-edges$weight), ]

    if (!is.null(top_n)) {
      stopifnot(is.numeric(top_n), top_n > 0)
      top_n <- as.integer(top_n)
      if (nrow(edges) > top_n) edges <- edges[seq_len(top_n), ]
    }
  }

  rownames(edges) <- NULL
  class(edges) <- c("cooccurrence", "data.frame")
  attr(edges, "matrix")         <- W_sparse
  attr(edges, "raw_matrix")     <- C_raw
  attr(edges, "items")          <- items
  attr(edges, "frequencies")    <- freq
  attr(edges, "similarity")     <- similarity
  attr(edges, "scale")          <- scale_method
  attr(edges, "threshold")      <- threshold
  attr(edges, "min_occur")      <- min_occur
  attr(edges, "n_transactions") <- n_trans
  attr(edges, "n_items")        <- n_items

  edges
}


# ---- Weighted core (long format with per-entity weights) ----

#' @noRd
.co_core_weighted <- function(data, field, by, weight_by, similarity,
                               scale_method, threshold, min_occur, top_n) {
  stopifnot(
    is.data.frame(data),
    field %in% names(data), by %in% names(data), weight_by %in% names(data)
  )

  docs      <- as.character(data[[by]])
  items_col <- as.character(data[[field]])
  weights   <- as.numeric(data[[weight_by]])

  ## Drop NA / zero-weight rows up front.
  keep <- !is.na(docs) & !is.na(items_col) & !is.na(weights) & weights != 0
  docs <- docs[keep]
  items_col <- items_col[keep]
  weights <- weights[keep]

  all_docs  <- unique(docs)
  all_items <- sort(unique(items_col))
  doc_idx   <- match(docs, all_docs)
  item_idx  <- match(items_col, all_items)

  ## Sparse weighted matrix: docs x items. Duplicate (doc, item) rows sum.
  W <- Matrix::sparseMatrix(
    i = doc_idx, j = item_idx, x = weights,
    dims = c(length(all_docs), length(all_items)),
    dimnames = list(NULL, all_items)
  )

  if (min_occur > 1L) {
    ## Count distinct docs per item via sparse triplets.
    n_docs_per_item <- tabulate(item_idx, nbins = length(all_items))
    keep_items <- n_docs_per_item >= min_occur
    W <- W[, keep_items, drop = FALSE]
    all_items <- all_items[keep_items]
  }

  if (ncol(W) == 0L)
    stop("No items remain after min_occur filtering.", call. = FALSE)

  n_trans <- nrow(W)
  n_items <- ncol(W)
  freq <- as.numeric(Matrix::colSums(W))
  names(freq) <- all_items

  ## Binary companion for raw counts.
  B_bin <- Matrix::sparseMatrix(
    i = doc_idx, j = item_idx, x = 1,
    dims = c(length(all_docs), length(all_items)),
    dimnames = list(NULL, colnames(W))
  )
  ## If min_occur filtered columns, realign B_bin to W.
  if (ncol(B_bin) != ncol(W)) {
    B_bin <- B_bin[, colnames(W), drop = FALSE]
  }
  ## sparseMatrix sums duplicates — clip to {0, 1} for the binary matrix.
  B_bin <- Matrix::drop0(sign(B_bin))

  C_raw <- Matrix::crossprod(B_bin)
  C     <- Matrix::crossprod(W)
  Matrix::diag(C) <- 0
  Matrix::diag(C_raw) <- 0
  C     <- Matrix::drop0(C)
  C_raw <- Matrix::drop0(C_raw)

  .co_finalize(C, C_raw, freq, n_trans, n_items, all_items,
               similarity, scale_method, threshold, min_occur, top_n)
}


# ---- Scaling (triplet-based) ----

#' Apply a post-normalisation scaling to edge weights.
#'
#' @param vals Numeric vector to scale (the edge weights returned to the user).
#' @param population Numeric vector of all non-zero values that would appear in
#'   the conceptual full k x k matrix. Used to compute statistics (min/max,
#'   mean/sd, sum) so results match the original dense implementation.
#' @param method Scaling method.
#' @noRd
.co_scale_values <- function(vals, population, method) {
  nz <- population[population != 0]
  if (length(nz) == 0L) return(vals)

  nonzero <- vals != 0

  switch(method,
    minmax = {
      mn <- min(nz); mx <- max(nz)
      out <- vals
      out[nonzero] <- if (mx > mn) (vals[nonzero] - mn) / (mx - mn) else 1
      out
    },
    log = {
      out <- vals
      out[nonzero] <- log(1 + vals[nonzero])
      out
    },
    log10 = {
      out <- vals
      out[nonzero] <- log10(1 + vals[nonzero])
      out
    },
    binary = {
      out <- vals
      out[nonzero] <- 1
      out
    },
    zscore = {
      mu <- mean(nz); s <- stats::sd(nz)
      out <- vals
      if (s > 0) out[nonzero] <- (vals[nonzero] - mu) / s
      out
    },
    sqrt = {
      out <- vals
      out[nonzero] <- sqrt(vals[nonzero])
      out
    },
    proportion = {
      s <- sum(nz)
      out <- vals
      if (s > 0) out[nonzero] <- vals[nonzero] / s
      out
    }
  )
}


# ---- Format detection ----

#' @noRd
.co_detect_format <- function(data, field, by, sep) {
  if (is.list(data) && !is.data.frame(data) && !is.matrix(data))
    return("list")

  if (!is.null(field) && length(field) == 1L && identical(field, "all"))
    return("wide")

  if (!is.null(sep) && !is.null(field) && length(field) > 1L)
    return("multi_delimited")

  if (!is.null(sep) && !is.null(field))
    return("delimited")

  if (!is.null(field) && !is.null(by))
    return("long")

  if (!is.null(field) && is.null(sep) && is.null(by) && is.data.frame(data))
    return("field_only")

  if (is.matrix(data) || is.data.frame(data)) {
    mat <- if (is.data.frame(data)) as.matrix(data) else data
    if (is.numeric(mat) && all(mat[!is.na(mat)] %in% c(0, 1)))
      return("binary")
    stop("Cannot detect input format. For wide sequence data use field = \"all\".",
         call. = FALSE)
  }

  stop("Cannot detect input format. Provide field/by/sep arguments or a ",
       "recognized data structure (data.frame, matrix, list).", call. = FALSE)
}


#' @noRd
.co_warn_missing_sep <- function(data, field) {
  candidates <- c(";", ",", "|", "/", "\t")
  vals <- unlist(lapply(field, function(f) as.character(data[[f]])))
  vals <- vals[!is.na(vals)]
  if (length(vals) == 0L) return(invisible())

  counts <- vapply(candidates, function(s) sum(grepl(s, vals, fixed = TRUE)),
                   integer(1))
  best <- which.max(counts)

  if (counts[best] > 0L) {
    sep_label <- if (candidates[best] == "\t") "\\t" else candidates[best]
    pct <- round(100 * counts[best] / length(vals))
    warning(
      "`field` was provided without `sep`. Each value is treated as a single item. ",
      sprintf(
        "Found '%s' in %d%% of values. Did you mean: sep = \"%s\"?",
        sep_label, pct, sep_label
      ),
      call. = FALSE
    )
  } else {
    warning(
      "`field` was provided without `sep`. Each value is treated as a single item.",
      call. = FALSE
    )
  }
}

# ---- Parsers ----

#' Rebuild a list of per-row character vectors from a flat token vector.
#'
#' Splits `flat` (already trimmed and NA-cleaned) back into a list of length
#' `n`, honouring the original row lengths via the accompanying `row_idx`.
#' Rows whose tokens were all dropped appear as `character(0)`. Per-row
#' deduplication is done globally via `duplicated()` on a two-column frame,
#' which is faster than calling `unique()` once per row.
#'
#' @noRd
.co_relist_unique <- function(flat, row_idx, n) {
  if (length(flat) == 0L) return(rep(list(character(0)), n))

  ## Global dedup: at most one (row, item) pair.
  dup <- duplicated(data.frame(row_idx, flat, stringsAsFactors = FALSE))
  flat <- flat[!dup]
  row_idx <- row_idx[!dup]

  ## split() by a factor with the full level range preserves empty rows.
  out <- split(flat, factor(row_idx, levels = seq_len(n)))
  names(out) <- NULL
  out
}

#' @noRd
.co_parse_delimited <- function(data, field, sep) {
  stopifnot(is.data.frame(data), length(field) == 1L, field %in% names(data))
  vals <- as.character(data[[field]])
  n <- length(vals)

  ## strsplit returns a list; flatten once so `trimws` and filtering run as
  ## vectorised C loops over the full token population rather than 166k
  ## per-row R calls (that per-row path was ~48% of total runtime).
  splits <- strsplit(vals, sep, fixed = TRUE)
  lens <- lengths(splits)
  if (sum(lens) == 0L) return(rep(list(character(0)), n))

  flat <- trimws(unlist(splits, use.names = FALSE))
  row_idx <- rep.int(seq_len(n), lens)
  keep <- !is.na(flat) & nzchar(flat)
  flat <- flat[keep]; row_idx <- row_idx[keep]

  .co_relist_unique(flat, row_idx, n)
}

#' @noRd
.co_parse_multi_delimited <- function(data, field, sep) {
  stopifnot(is.data.frame(data), all(field %in% names(data)))
  n <- nrow(data)

  ## Split each column independently, concatenate with aligned row indices,
  ## then run the same flat trim + dedup as `.co_parse_delimited`.
  parts <- lapply(field, function(f) {
    splits <- strsplit(as.character(data[[f]]), sep, fixed = TRUE)
    lens <- lengths(splits)
    list(
      row_idx = rep.int(seq_len(n), lens),
      flat    = unlist(splits, use.names = FALSE)
    )
  })
  row_idx <- unlist(lapply(parts, `[[`, "row_idx"), use.names = FALSE)
  flat    <- unlist(lapply(parts, `[[`, "flat"),    use.names = FALSE)
  if (length(flat) == 0L) return(rep(list(character(0)), n))

  flat <- trimws(flat)
  keep <- !is.na(flat) & nzchar(flat)
  flat <- flat[keep]; row_idx <- row_idx[keep]

  .co_relist_unique(flat, row_idx, n)
}

#' @noRd
.co_parse_field_only <- function(data, field) {
  stopifnot(is.data.frame(data), all(field %in% names(data)))
  lapply(seq_len(nrow(data)), function(i) {
    vals <- as.character(unlist(data[i, field, drop = TRUE]))
    vals <- trimws(vals)
    vals <- vals[!is.na(vals) & nzchar(vals)]
    unique(vals)
  })
}

#' @noRd
.co_parse_long <- function(data, field, by) {
  stopifnot(is.data.frame(data), field %in% names(data), by %in% names(data))
  groups <- split(as.character(data[[field]]), data[[by]])
  lapply(groups, function(items) {
    items <- items[nzchar(items) & !is.na(items)]
    unique(items)
  })
}

#' @noRd
.co_parse_binary <- function(data) {
  mat <- if (is.data.frame(data)) as.matrix(data) else data
  if (is.null(colnames(mat)))
    colnames(mat) <- paste0("V", seq_len(ncol(mat)))
  cn <- colnames(mat)
  lapply(seq_len(nrow(mat)), function(i) cn[mat[i, ] == 1])
}

#' @noRd
.co_parse_wide <- function(data) {
  mat <- if (is.data.frame(data)) as.matrix(data) else data
  lapply(seq_len(nrow(mat)), function(i) {
    vals <- as.character(mat[i, ])
    vals <- vals[!is.na(vals) & nzchar(vals) & !(vals %in% .void_markers)]
    unique(vals)
  })
}

#' @noRd
.co_parse_list <- function(data) {
  lapply(data, function(items) {
    items <- as.character(items)
    items <- items[!is.na(items) & nzchar(items)]
    unique(items)
  })
}


# ---- Sparse bipartite matrix builder ----

#' Build sparse works-by-items incidence matrices from a list of transactions.
#'
#' Returns both the counting-weighted matrix `B` (used for the weighted
#' crossprod) and the binary matrix `B_bin` (used for item frequencies and
#' raw counts). Staying in sparse representation is what lets the pipeline
#' scale to hundreds of thousands of items.
#'
#' Counting method `"fractional"` (Perianes-Rodriguez et al., 2016) bakes
#' `sqrt(1/(n_r - 1))` into each row's non-zero entries, so that
#' `crossprod(B)[i, j] = sum over rows with both i and j of 1/(n_r - 1)`.
#'
#' @noRd
.co_build_sparse <- function(transactions, counting) {
  all_items <- sort(unique(unlist(transactions, use.names = FALSE)))
  n <- length(transactions)
  k <- length(all_items)
  lens <- vapply(transactions, length, integer(1))
  row_idx <- rep.int(seq_len(n), lens)
  col_idx <- match(unlist(transactions, use.names = FALSE), all_items)

  if (counting == "fractional") {
    row_weight <- ifelse(lens > 1L, 1 / (lens - 1L), 1)
    x <- sqrt(row_weight[row_idx])
  } else {
    x <- rep(1.0, length(row_idx))
  }

  B <- Matrix::sparseMatrix(
    i = row_idx, j = col_idx, x = x,
    dims = c(n, k),
    dimnames = list(NULL, all_items)
  )
  B_bin <- Matrix::sparseMatrix(
    i = row_idx, j = col_idx, x = rep(1.0, length(row_idx)),
    dims = c(n, k),
    dimnames = list(NULL, all_items)
  )

  list(B = B, B_bin = B_bin, items = all_items, n = n, k = k)
}


# ---- Output format conversion ----

#' @noRd
.co_format_output <- function(result, output) {
  if (output == "default") return(result)

  if (output == "gephi") {
    out <- result
    has_group <- "group" %in% names(out)
    names(out)[names(out) == "from"] <- "Source"
    names(out)[names(out) == "to"] <- "Target"
    names(out)[names(out) == "weight"] <- "Weight"
    names(out)[names(out) == "count"] <- "Count"
    out$Type <- "Undirected"
    # Reorder: Source, Target, Weight, Type, Count, [group]
    cols <- c("Source", "Target", "Weight", "Type", "Count")
    if (has_group) cols <- c(cols, "group")
    out <- out[, cols]
    class(out) <- c("cooccurrence", "data.frame")
    # Copy attributes
    for (a in c("matrix", "raw_matrix", "items", "frequencies",
                "similarity", "scale", "threshold", "min_occur",
                "n_transactions", "n_items", "split_by", "groups")) {
      attr(out, a) <- attr(result, a)
    }
    return(out)
  }

  if (output == "igraph") return(as_igraph(result))
  if (output == "cograph") return(as_cograph(result))
  if (output == "matrix") return(as_matrix(result))

  result
}

# ---- S3 methods for cooccurrence ----

#' Print a cooccurrence edge list
#'
#' @param x A \code{cooccurrence} data frame.
#' @param n Integer. Number of rows to show. Default 10.
#' @param ... Ignored.
#' @return Invisibly returns \code{x}.
#' @examples
#' res <- cooccurrence(list(c("A","B","C"), c("B","C"), c("A","C")))
#' print(res)
#' @export
print.cooccurrence <- function(x, n = 10L, ...) {
  nodes <- length(unique(c(x$from, x$to)))
  edges <- nrow(x)
  sim <- attr(x, "similarity")
  sc <- attr(x, "scale")
  n_trans <- attr(x, "n_transactions")

  sb <- attr(x, "split_by")
  grps <- attr(x, "groups")

  cat(sprintf("# cooccurrence: %d nodes, %d edges", nodes, edges))
  if (!is.null(n_trans)) cat(sprintf(" (%d transactions)", n_trans))
  if (!is.null(sb)) cat(sprintf(" | split_by: %s (%d groups)", sb, length(grps)))
  if (!is.null(sim) && sim != "none") cat(sprintf(" | similarity: %s", sim))
  if (!is.null(sc) && sc != "none") cat(sprintf(" | scale: %s", sc))
  cat("\n")

  show <- min(n, edges)
  if (show > 0L) {
    print(as.data.frame(x)[seq_len(show), ], row.names = FALSE)
    if (edges > show)
      cat(sprintf("# ... %d more edges\n", edges - show))
  } else {
    cat("# (no edges)\n")
  }
  invisible(x)
}


#' Summarise a cooccurrence network
#'
#' @param object A \code{cooccurrence} data frame.
#' @param ... Ignored.
#' @return Invisibly returns \code{object}.
#' @examples
#' res <- cooccurrence(list(c("A","B","C"), c("B","C"), c("A","C")))
#' summary(res)
#' @export
summary.cooccurrence <- function(object, ...) {
  nodes <- unique(c(object$from, object$to))
  n_nodes <- length(nodes)
  n_edges <- nrow(object)
  max_possible <- n_nodes * (n_nodes - 1L) / 2L
  density <- if (max_possible > 0) n_edges / max_possible else NA_real_

  cat("cooccurrence network\n")
  cat(rep("-", 30), "\n", sep = "")
  cat(sprintf("Nodes          : %d\n", n_nodes))
  cat(sprintf("Edges          : %d\n", n_edges))
  cat(sprintf("Density        : %.4f\n", density))
  cat(sprintf("Transactions   : %d\n", attr(object, "n_transactions")))
  cat(sprintf("Similarity     : %s\n", attr(object, "similarity")))

  sc <- attr(object, "scale")
  if (!is.null(sc) && sc != "none")
    cat(sprintf("Scale          : %s\n", sc))

  if (n_edges > 0L) {
    cat(sprintf("Weight range   : [%.4g, %.4g]\n",
                min(object$weight), max(object$weight)))
    cat(sprintf("Count range    : [%d, %d]\n",
                min(object$count), max(object$count)))

    # Top 5 nodes by degree
    deg <- sort(table(c(object$from, object$to)), decreasing = TRUE)
    top <- utils::head(deg, 5L)
    cat(sprintf("Top nodes      : %s\n",
                paste(sprintf("%s(%d)", names(top), as.integer(top)),
                      collapse = ", ")))
  }

  invisible(object)
}


#' Plot a cooccurrence network
#'
#' Plots the co-occurrence matrix as a heatmap. If \pkg{igraph} is available,
#' plots a network graph instead.
#'
#' @param x A \code{cooccurrence} data frame.
#' @param type Character. \code{"heatmap"} (default) or \code{"network"}
#'   (requires \pkg{igraph}).
#' @param ... Passed to the plotting function.
#' @return Invisibly returns \code{x}.
#' @examples
#' res <- cooccurrence(list(c("A","B","C"), c("B","C"), c("A","C")))
#' plot(res)
#' @export
plot.cooccurrence <- function(x, type = c("heatmap", "network"), ...) {
  type <- match.arg(type)

  if (type == "network") {
    if (!requireNamespace("igraph", quietly = TRUE))
      stop("Package 'igraph' is required for network plots.", call. = FALSE)
    g <- as_igraph(x)
    plot(g, ...)
  } else {
    ## Go through as_matrix() so sparse attributes get densified consistently.
    mat <- as_matrix(x)
    n <- nrow(mat)
    graphics::image(
      seq_len(n), seq_len(n), t(mat[n:1, ]),
      xlab = "", ylab = "", axes = FALSE, ...
    )
    graphics::axis(1, at = seq_len(n), labels = colnames(mat),
                   las = 2, cex.axis = 0.7)
    graphics::axis(2, at = seq_len(n), labels = rev(rownames(mat)),
                   las = 2, cex.axis = 0.7)
  }

  invisible(x)
}

#' Launch the cooccure Shiny explorer
#'
#' Opens an interactive Shiny application for building and exploring
#' co-occurrence networks. Requires the \pkg{shiny} and \pkg{DT} packages.
#'
#' @param ... Passed to \code{\link[shiny]{runApp}} (e.g. \code{port},
#'   \code{launch.browser}).
#' @return Called for its side effect (launches the app). No return value.
#' @export
#' @examples
#' if (interactive()) {
#'   launch_app()
#' }
launch_app <- function(...) {
  # nocov start — interactive Shiny entrypoint, not reachable under batch test runs
  for (pkg in c("shiny", "DT")) {
    if (!requireNamespace(pkg, quietly = TRUE))
      stop(sprintf("Package '%s' is required. Install it with: install.packages('%s')",
                   pkg, pkg), call. = FALSE)
  }
  app_dir <- system.file("shiny", "cooccure_app", package = "cooccure")
  if (!nzchar(app_dir))
    stop("Could not find the Shiny app directory. Re-install the cooccure package.",
         call. = FALSE)
  shiny::runApp(app_dir, ...)
  # nocov end
}

# ---- Package datasets ----

#' IMDB movie metadata (1970-2024)
#'
#' A sample of 1,000 highly-rated IMDB movies (rating >= 7.0, >= 1,000 votes)
#' released between 1970 and 2024. The \code{genres} column is comma-delimited
#' and suitable for use as the \code{field} argument to \code{\link{cooccurrence}}.
#'
#' @format A data frame with 1,000 rows and 7 variables:
#' \describe{
#'   \item{tconst}{IMDB title identifier (e.g. \code{"tt0068646"}).}
#'   \item{primaryTitle}{Movie title.}
#'   \item{startYear}{Release year (integer).}
#'   \item{genres}{Comma-separated genre labels (e.g. \code{"Crime,Drama"}).}
#'   \item{decade}{Release decade as a character string (e.g. \code{"1970s"}).}
#'   \item{averageRating}{IMDB average user rating.}
#'   \item{numVotes}{Number of IMDB user votes.}
#' }
#' @source \url{https://developer.imdb.com/non-commercial-datasets/}
#' @examples
#' head(movies)
#' cooccurrence(movies, field = "genres", sep = ",", similarity = "jaccard")
"movies"


#' IMDB actor-movie long table (1970-2024)
#'
#' Long-format bipartite table linking actors to movies in
#' \code{\link{movies}}. Pre-filtered to the 624 actors who appear in at
#' least two movies, so all similarity measures compute instantly.
#' Pass \code{field = "actor"} and \code{by = "tconst"} to
#' \code{\link{cooccurrence}} to build an actor co-appearance network.
#'
#' @format A data frame with 1,267 rows and 7 variables:
#' \describe{
#'   \item{actor}{Actor name.}
#'   \item{tconst}{IMDB title identifier linking to \code{\link{movies}}.}
#'   \item{primaryTitle}{Movie title.}
#'   \item{startYear}{Release year (integer).}
#'   \item{decade}{Release decade as a character string.}
#'   \item{genres}{Comma-separated genre labels for the linked movie.}
#'   \item{averageRating}{IMDB average user rating for the linked movie.}
#' }
#' @source \url{https://developer.imdb.com/non-commercial-datasets/}
#' @examples
#' head(actors)
#' cooccurrence(actors, field = "actor", by = "tconst", similarity = "jaccard")
"actors"


#' IMDB actor-genre long table (1970-2024)
#'
#' Long-format table mapping each of the 624 actors in \code{\link{actors}}
#' to every genre of every movie they appeared in. Use this to build an
#' actor co-occurrence network grouped by genre: which actors share the
#' same genres? Pass \code{field = "actor"} and \code{by = "genre"} to
#' \code{\link{cooccurrence}}.
#'
#' @format A data frame with 2,502 rows and 2 variables:
#' \describe{
#'   \item{actor}{Actor name.}
#'   \item{genre}{Genre label (one row per actor-genre combination).}
#' }
#' @source \url{https://developer.imdb.com/non-commercial-datasets/}
#' @examples
#' head(actor_genres)
#' cooccurrence(actor_genres, field = "actor", by = "genre", similarity = "jaccard")
"actor_genres"


#' Demo actor-movie-genre table
#'
#' A small hand-crafted dataset of 30 well-known actors across 10 classic
#' films with genre labels. Designed for quick exploration in the Shiny app.
#' Use \code{field = "actor"} with \code{by = "movie"} or \code{by = "genre"}.
#'
#' @format A data frame with 34 rows and 3 variables:
#' \describe{
#'   \item{movie}{Movie title.}
#'   \item{actor}{Actor name.}
#'   \item{genre}{Primary genre label.}
#' }
#' @examples
#' head(demo)
#' cooccurrence(demo, field = "actor", by = "movie", similarity = "jaccard")
"demo"

# ---- Converters for cooccurrence objects ----

#' Extract the co-occurrence matrix
#'
#' Returns the full square co-occurrence matrix (normalized + scaled).
#' Use \code{type = "raw"} for the raw count matrix.
#'
#' @param x A \code{cooccurrence} data frame.
#' @param type Character. \code{"normalized"} (default) or \code{"raw"}.
#' @param ... Ignored.
#' @return A numeric matrix.
#' @examples
#' res <- cooccurrence(list(c("A","B","C"), c("B","C"), c("A","C")))
#' as_matrix(res)
#' as_matrix(res, type = "raw")
#' @export
as_matrix <- function(x, ...) UseMethod("as_matrix")

#' @rdname as_matrix
#' @export
as_matrix.cooccurrence <- function(x, type = c("normalized", "raw"), ...) {
  type <- match.arg(type)
  mat <- if (type == "raw") attr(x, "raw_matrix") else attr(x, "matrix")
  if (!is.null(mat)) {
    ## Attributes are stored as sparse Matrix objects in the current engine.
    ## Densify on demand so downstream code gets a regular base matrix. For
    ## very large networks, the caller should avoid as_matrix() and work with
    ## the edge list directly.
    if (inherits(mat, "Matrix")) return(as.matrix(mat))
    return(mat)
  }

  ## Fallback: rebuild from the edge list (used only if attribute is missing).
  items <- sort(unique(c(x$from, x$to)))
  k <- length(items)
  col <- if (type == "raw") "count" else "weight"
  if (nrow(x) == 0L) {
    return(matrix(0, k, k, dimnames = list(items, items)))
  }
  i <- match(x$from, items)
  j <- match(x$to, items)
  vals <- x[[col]]
  M <- matrix(0, k, k, dimnames = list(items, items))
  M[cbind(i, j)] <- vals
  M[cbind(j, i)] <- vals
  M
}


#' Convert to igraph
#'
#' Creates an undirected, weighted \code{igraph} graph from a
#' \code{cooccurrence} edge list.
#'
#' @param x A \code{cooccurrence} data frame.
#' @param ... Passed to \code{igraph::graph_from_data_frame}.
#' @return An \code{igraph} object.
#' @examples
#' res <- cooccurrence(list(c("A","B","C"), c("B","C"), c("A","C")))
#' if (requireNamespace("igraph", quietly = TRUE)) {
#'   g <- as_igraph(res)
#'   igraph::vcount(g)
#' }
#' @export
as_igraph <- function(x, ...) UseMethod("as_igraph")

#' @rdname as_igraph
#' @export
as_igraph.cooccurrence <- function(x, ...) {
  if (!requireNamespace("igraph", quietly = TRUE))
    stop("Package 'igraph' is required.", call. = FALSE)

  items <- attr(x, "items")
  vertices <- if (!is.null(items)) {
    data.frame(name = items, stringsAsFactors = FALSE)
  } else {
    NULL
  }

  igraph::graph_from_data_frame(
    x[, c("from", "to", "weight", "count")],
    directed = FALSE,
    vertices = vertices,
    ...
  )
}


#' Convert to tidygraph
#'
#' Creates a \code{tbl_graph} from a \code{cooccurrence} edge list.
#'
#' @param x A \code{cooccurrence} data frame.
#' @param ... Ignored.
#' @return A \code{tbl_graph} object.
#' @examples
#' res <- cooccurrence(list(c("A","B","C"), c("B","C"), c("A","C")))
#' if (requireNamespace("tidygraph", quietly = TRUE) &&
#'     requireNamespace("igraph",    quietly = TRUE)) {
#'   as_tidygraph(res)
#' }
#' @export
as_tidygraph <- function(x, ...) UseMethod("as_tidygraph")

#' @rdname as_tidygraph
#' @export
as_tidygraph.cooccurrence <- function(x, ...) {
  if (!requireNamespace("tidygraph", quietly = TRUE))
    stop("Package 'tidygraph' is required.", call. = FALSE)
  tidygraph::as_tbl_graph(as_igraph(x, ...))
}


#' Convert to cograph network
#'
#' Creates a \code{cograph_network} object from a \code{cooccurrence}
#' edge list, compatible with \code{cograph::splot()} and other cograph
#' functions.
#'
#' @param x A \code{cooccurrence} data frame.
#' @param ... Ignored.
#' @return A \code{cograph_network} object.
#' @examples
#' res <- cooccurrence(list(c("A","B","C"), c("B","C"), c("A","C")))
#' if (requireNamespace("cograph", quietly = TRUE)) {
#'   net <- as_cograph(res)
#'   net$n_nodes
#' }
#' @export
as_cograph <- function(x, ...) UseMethod("as_cograph")

#' @rdname as_cograph
#' @export
as_cograph.cooccurrence <- function(x, ...) {
  if (!requireNamespace("cograph", quietly = TRUE))
    stop("Package 'cograph' is required.", call. = FALSE)

  mat <- as_matrix(x, type = "normalized")
  items <- colnames(mat)

  nodes_df <- data.frame(
    id = seq_along(items), label = items, name = items,
    x = NA_real_, y = NA_real_, stringsAsFactors = FALSE
  )

  # Edge list from upper triangle
  idx <- which(upper.tri(mat) & mat != 0, arr.ind = TRUE)
  if (nrow(idx) > 0L) {
    edges_df <- data.frame(
      from = as.integer(idx[, 1]),
      to = as.integer(idx[, 2]),
      weight = mat[idx],
      stringsAsFactors = FALSE
    )
  } else {
    edges_df <- data.frame(from = integer(0), to = integer(0),
                           weight = numeric(0), stringsAsFactors = FALSE)
  }

  structure(
    list(
      weights = mat,
      nodes = nodes_df,
      edges = edges_df,
      directed = FALSE,
      n_nodes = length(items),
      n_edges = nrow(edges_df),
      meta = list(source = "cooccure", layout = NULL)
    ),
    class = "cograph_network"
  )
}


#' Convert to Nestimate netobject
#'
#' Creates a \code{netobject} from a \code{cooccurrence} edge list,
#' compatible with \code{Nestimate::centrality()},
#' \code{Nestimate::bootstrap_network()}, etc.
#'
#' @param x A \code{cooccurrence} data frame.
#' @param ... Ignored.
#' @return A \code{netobject} with class \code{c("netobject", "cograph_network")}.
#' @examples
#' res <- cooccurrence(list(c("A","B","C"), c("B","C"), c("A","C")))
#' if (requireNamespace("Nestimate", quietly = TRUE)) {
#'   net <- as_netobject(res)
#'   net$n_nodes
#' }
#' @export
as_netobject <- function(x, ...) UseMethod("as_netobject")

#' @rdname as_netobject
#' @export
as_netobject.cooccurrence <- function(x, ...) {
  if (!requireNamespace("Nestimate", quietly = TRUE))
    stop("Package 'Nestimate' is required.", call. = FALSE)

  mat <- as_matrix(x, type = "normalized")
  items <- colnames(mat)

  nodes_df <- data.frame(
    id = seq_along(items), label = items, name = items,
    x = NA_real_, y = NA_real_, stringsAsFactors = FALSE
  )

  idx <- which(upper.tri(mat) & mat != 0, arr.ind = TRUE)
  if (nrow(idx) > 0L) {
    edges_df <- data.frame(
      from = as.integer(idx[, 1]),
      to = as.integer(idx[, 2]),
      weight = mat[idx],
      stringsAsFactors = FALSE
    )
  } else {
    edges_df <- data.frame(from = integer(0), to = integer(0),
                           weight = numeric(0), stringsAsFactors = FALSE)
  }

  structure(
    list(
      data = NULL,
      weights = mat,
      nodes = nodes_df,
      edges = edges_df,
      directed = FALSE,
      method = "cooccurrence",
      params = list(
        similarity = attr(x, "similarity"),
        scale = attr(x, "scale"),
        threshold = attr(x, "threshold"),
        n_transactions = attr(x, "n_transactions")
      ),
      scaling = NULL,
      threshold = attr(x, "threshold"),
      n_nodes = length(items),
      n_edges = nrow(edges_df),
      level = NULL,
      meta = list(source = "cooccure", layout = NULL,
                  tna = list(method = "cooccurrence")),
      node_groups = NULL
    ),
    class = c("netobject", "cograph_network")
  )
}

# ---- cooccurrence() tests ----

.test_list <- list(c("A", "B", "C"), c("B", "C"), c("A", "C"))

# ========================================
# 1. Returns tidy data.frame
# ========================================

test_that("cooccurrence returns a tidy data.frame", {
  res <- cooccurrence(.test_list)
  expect_s3_class(res, "cooccurrence")
  expect_s3_class(res, "data.frame")
  expect_true(all(c("from", "to", "weight", "count") %in% names(res)))
  expect_true(nrow(res) > 0)
})

test_that("co() alias works", {
  res <- co(.test_list)
  expect_s3_class(res, "cooccurrence")
})

test_that("edges are sorted by weight descending", {
  res <- cooccurrence(.test_list)
  expect_true(all(diff(res$weight) <= 0))
})

# ========================================
# 2. Input formats
# ========================================

test_that("delimited input works", {
  df <- data.frame(items = c("A;B;C", "B;C", "A;C"), stringsAsFactors = FALSE)
  res <- cooccurrence(df, field = "items", sep = ";")
  expect_equal(nrow(res), 3L)  # 3 pairs: A-B, A-C, B-C
})

test_that("multi-column delimited works", {
  df <- data.frame(c1 = c("A;B", "B", "A"), c2 = c("C", "C", "C"),
                   stringsAsFactors = FALSE)
  res <- cooccurrence(df, field = c("c1", "c2"), sep = ";")
  expect_equal(nrow(res), 3L)
})

test_that("long/bipartite works", {
  df <- data.frame(doc = c(1,1,1,2,2,3,3),
                   item = c("A","B","C","B","C","A","C"),
                   stringsAsFactors = FALSE)
  res <- cooccurrence(df, field = "item", by = "doc")
  expect_equal(nrow(res), 3L)
})

test_that("binary matrix works", {
  bin <- matrix(c(1,1,1, 0,1,1, 1,0,1), nrow = 3, byrow = TRUE,
                dimnames = list(NULL, c("A","B","C")))
  res <- cooccurrence(bin)
  expect_equal(nrow(res), 3L)
})

test_that("wide sequence works with field = 'all'", {
  df <- data.frame(V1 = c("A","B","A"), V2 = c("B","C","C"),
                   V3 = c("C", NA, NA), stringsAsFactors = FALSE)
  res <- cooccurrence(df, field = "all")
  expect_equal(nrow(res), 3L)
})

test_that("non-binary data frame without field = 'all' errors", {
  df <- data.frame(V1 = c("A","B","A"), V2 = c("B","C","C"),
                   stringsAsFactors = FALSE)
  expect_error(cooccurrence(df), "field = \"all\"")
})

test_that("list input works", {
  res <- cooccurrence(.test_list)
  expect_equal(nrow(res), 3L)
})

test_that("all formats give same counts", {
  res_list <- cooccurrence(.test_list)

  df_del <- data.frame(items = c("A;B;C", "B;C", "A;C"),
                        stringsAsFactors = FALSE)
  res_del <- cooccurrence(df_del, field = "items", sep = ";")

  # Sort both by from+to for comparison
  key <- function(r) paste(pmin(r$from, r$to), pmax(r$from, r$to))
  r1 <- res_list[order(key(res_list)), ]
  r2 <- res_del[order(key(res_del)), ]

  expect_equal(r1$count, r2$count)
  expect_equal(r1$weight, r2$weight)
})

# ========================================
# 3. Similarity measures
# ========================================

test_that("similarity = 'none' gives raw counts", {
  res <- cooccurrence(.test_list, similarity = "none")
  # A-C co-occur 2 times, so weight = 2
  ac <- res[res$from == "A" & res$to == "C", ]
  expect_equal(ac$weight, 2)
  expect_equal(ac$count, 2L)
})

test_that("similarity = 'jaccard' is correct", {
  res <- cooccurrence(.test_list, similarity = "jaccard")
  ab <- res[res$from == "A" & res$to == "B", ]
  # Jaccard(A,B) = 1 / (2+2-1) = 1/3
  expect_equal(ab$weight, 1 / 3)
})

test_that("similarity = 'cosine' is correct", {
  res <- cooccurrence(.test_list, similarity = "cosine")
  ab <- res[res$from == "A" & res$to == "B", ]
  expect_equal(ab$weight, 0.5)
})

test_that("similarity = 'association' is correct", {
  res <- cooccurrence(.test_list, similarity = "association")
  ab <- res[res$from == "A" & res$to == "B", ]
  expect_equal(ab$weight, 0.25)
})

test_that("similarity = 'equivalence' is correct", {
  res <- cooccurrence(.test_list, similarity = "equivalence")
  ab <- res[res$from == "A" & res$to == "B", ]
  expect_equal(ab$weight, 0.25)
})

test_that("similarity = 'dice' is correct", {
  res <- cooccurrence(.test_list, similarity = "dice")
  ab <- res[res$from == "A" & res$to == "B", ]
  expect_equal(ab$weight, 0.5)
})

test_that("similarity = 'inclusion' is correct", {
  res <- cooccurrence(.test_list, similarity = "inclusion")
  ac <- res[res$from == "A" & res$to == "C", ]
  expect_equal(ac$weight, 1)
})

test_that("similarity = 'relative' row-normalizes", {
  res <- cooccurrence(.test_list, similarity = "relative")
  mat <- as_matrix(res)
  expect_true(all(abs(rowSums(mat) - 1) < 1e-10 | rowSums(mat) == 0))
})

# ========================================
# 4. Scaling
# ========================================

test_that("scale = 'minmax' scales to [0,1]", {
  res <- cooccurrence(.test_list, scale = "minmax")
  expect_true(max(res$weight) <= 1)
  expect_true(min(res$weight) >= 0)
})

test_that("scale = 'log' applies log(1+w)", {
  res_raw <- cooccurrence(.test_list)
  res_log <- cooccurrence(.test_list, scale = "log")
  expect_equal(res_log$weight, log(1 + res_raw$weight))
})

test_that("scale = 'binary' gives 0/1 weights", {
  res <- cooccurrence(.test_list, scale = "binary")
  expect_true(all(res$weight %in% c(0, 1)))
})

test_that("scale = 'sqrt' applies square root", {
  res_raw <- cooccurrence(.test_list)
  res_sqrt <- cooccurrence(.test_list, scale = "sqrt")
  expect_equal(res_sqrt$weight, sqrt(res_raw$weight))
})

test_that("scale = 'zscore' standardizes", {
  res <- cooccurrence(.test_list, scale = "zscore")
  vals <- res$weight[res$weight != 0]
  if (length(vals) > 1) {
    expect_equal(mean(vals), 0, tolerance = 1e-10)
  }
})

test_that("scale = 'proportion' normalizes by total weight", {
  res <- cooccurrence(.test_list, scale = "proportion")
  # The full symmetric matrix sums to 1 (upper + lower triangle)
  mat <- as_matrix(res)
  expect_equal(sum(mat), 1, tolerance = 1e-10)
})

# ========================================
# 5. Parameters
# ========================================

test_that("threshold filters edges", {
  res <- cooccurrence(.test_list, threshold = 2)
  expect_true(all(res$weight >= 2))
  expect_equal(nrow(res), 2L)  # Only A-C and B-C have count >= 2
})

test_that("min_occur drops rare entities", {
  trans <- list(c("A","B"), c("A","B"), c("A","D"))
  res <- cooccurrence(trans, min_occur = 2L)
  expect_false("D" %in% c(res$from, res$to))
})

test_that("top_n limits edges", {
  res <- cooccurrence(.test_list, top_n = 1L)
  expect_equal(nrow(res), 1L)
})

# ========================================
# 6. Attributes
# ========================================

test_that("attributes are stored", {
  res <- cooccurrence(.test_list, similarity = "jaccard", scale = "log")
  expect_equal(attr(res, "similarity"), "jaccard")
  expect_equal(attr(res, "scale"), "log")
  expect_equal(attr(res, "n_transactions"), 3L)
  expect_equal(attr(res, "n_items"), 3L)
  expect_true(!is.null(attr(res, "matrix")))
  expect_true(!is.null(attr(res, "items")))
  expect_true(!is.null(attr(res, "frequencies")))
})

# ========================================
# 7. Converters
# ========================================

test_that("as_matrix returns correct matrix", {
  res <- cooccurrence(.test_list)
  mat <- as_matrix(res)
  expect_true(is.matrix(mat))
  expect_equal(nrow(mat), 3L)
  expect_equal(ncol(mat), 3L)
  expect_true(isSymmetric(mat))
})

test_that("as_matrix type='raw' gives counts", {
  res <- cooccurrence(.test_list, similarity = "jaccard")
  raw <- as_matrix(res, type = "raw")
  expect_equal(raw["A", "B"], 1)
  expect_equal(raw["A", "C"], 2)
})

test_that("as_igraph works", {
  skip_if_not_installed("igraph")
  res <- cooccurrence(.test_list)
  g <- as_igraph(res)
  expect_true(igraph::is_igraph(g))
  expect_equal(igraph::vcount(g), 3L)
  expect_false(igraph::is_directed(g))
})

test_that("as_cograph works", {
  skip_if_not_installed("cograph")
  res <- cooccurrence(.test_list)
  net <- as_cograph(res)
  expect_true(inherits(net, "cograph_network"))
  expect_equal(net$n_nodes, 3L)
})

test_that("as_netobject works", {
  skip_if_not_installed("Nestimate")
  res <- cooccurrence(.test_list)
  net <- as_netobject(res)
  expect_true(inherits(net, "netobject"))
  expect_true(inherits(net, "cograph_network"))
  expect_false(net$directed)
})

test_that("as_tidygraph works", {
  skip_if_not_installed("tidygraph")
  skip_if_not_installed("igraph")
  res <- cooccurrence(.test_list)
  tg <- as_tidygraph(res)
  expect_true(inherits(tg, "tbl_graph"))
})

# ========================================
# 8. Edge cases
# ========================================

test_that("error on empty input", {
  expect_error(cooccurrence(list()), "No non-empty transactions")
})

test_that("single-item transactions give no edges", {
  res <- cooccurrence(list(c("A"), c("B"), c("C")))
  expect_equal(nrow(res), 0L)
})

test_that("count is always the raw co-occurrence", {
  res <- cooccurrence(.test_list, similarity = "jaccard", scale = "log")
  ab <- res[res$from == "A" & res$to == "B", ]
  expect_equal(ab$count, 1L)
})

# ========================================
# 9. split_by
# ========================================

test_that("split_by produces a group column", {
  df <- data.frame(
    year = c(2020, 2020, 2020, 2021, 2021, 2021),
    kw = c("A; B; C", "B; C", "A; C", "B; C; D", "C; D", "B; D"),
    stringsAsFactors = FALSE
  )
  res <- cooccurrence(df, field = "kw", sep = ";", split_by = "year")
  expect_s3_class(res, "cooccurrence")
  expect_true("group" %in% names(res))
  expect_equal(sort(unique(res$group)), c("2020", "2021"))
})

test_that("split_by computes separate networks per group", {
  df <- data.frame(
    grp = c("X", "X", "Y", "Y"),
    kw = c("A; B", "A; B", "C; D", "C; D"),
    stringsAsFactors = FALSE
  )
  res <- cooccurrence(df, field = "kw", sep = ";", split_by = "grp")
  # Group X has only A-B; group Y has only C-D
  res_x <- res[res$group == "X", ]
  res_y <- res[res$group == "Y", ]
  expect_equal(nrow(res_x), 1L)
  expect_equal(nrow(res_y), 1L)
  expect_true(all(c(res_x$from, res_x$to) %in% c("A", "B")))
  expect_true(all(c(res_y$from, res_y$to) %in% c("C", "D")))
})

test_that("split_by + similarity works", {
  df <- data.frame(
    grp = c("X", "X", "X", "Y", "Y", "Y"),
    kw = c("A; B; C", "B; C", "A; C", "A; B; C", "B; C", "A; C"),
    stringsAsFactors = FALSE
  )
  res <- cooccurrence(df, field = "kw", sep = ";",
                      split_by = "grp", similarity = "jaccard")
  # Both groups have the same data → same weights
  res_x <- res[res$group == "X", ]
  res_y <- res[res$group == "Y", ]
  expect_equal(sort(res_x$weight), sort(res_y$weight))
})

test_that("split_by + top_n applies per group", {
  df <- data.frame(
    grp = c("X", "X", "X", "Y", "Y", "Y"),
    kw = c("A; B; C", "B; C", "A; C", "D; E; F", "E; F", "D; F"),
    stringsAsFactors = FALSE
  )
  res <- cooccurrence(df, field = "kw", sep = ";",
                      split_by = "grp", top_n = 1L)
  # 1 edge per group → 2 total
  expect_equal(nrow(res), 2L)
  expect_equal(length(unique(res$group)), 2L)
})

test_that("split_by skips groups with no edges", {
  df <- data.frame(
    grp = c("X", "X", "Y"),
    kw = c("A; B", "A; B", "C"),
    stringsAsFactors = FALSE
  )
  res <- cooccurrence(df, field = "kw", sep = ";", split_by = "grp")
  # Y has only one item → no edges → should be absent
  expect_equal(unique(res$group), "X")
})

test_that("split_by stores attributes", {
  df <- data.frame(
    year = c(2020, 2021),
    kw = c("A; B", "A; B"),
    stringsAsFactors = FALSE
  )
  res <- cooccurrence(df, field = "kw", sep = ";", split_by = "year")
  expect_equal(attr(res, "split_by"), "year")
  expect_equal(sort(attr(res, "groups")), c("2020", "2021"))
})

# ========================================
# 10. Output formats
# ========================================

test_that("output = 'gephi' renames columns", {
  res <- cooccurrence(.test_list, output = "gephi")
  expect_true(all(c("Source", "Target", "Weight", "Type", "Count") %in% names(res)))
  expect_false("from" %in% names(res))
  expect_equal(unique(res$Type), "Undirected")
})

test_that("output = 'gephi' with split_by includes group", {
  df <- data.frame(grp = c("X", "X", "Y", "Y"),
                   kw = c("A; B", "A; B", "C; D", "C; D"),
                   stringsAsFactors = FALSE)
  res <- cooccurrence(df, field = "kw", sep = ";",
                      split_by = "grp", output = "gephi")
  expect_true("group" %in% names(res))
  expect_true("Source" %in% names(res))
})

test_that("output = 'matrix' returns a matrix", {
  res <- cooccurrence(.test_list, output = "matrix")
  expect_true(is.matrix(res))
  expect_equal(nrow(res), 3L)
})

test_that("output = 'igraph' returns igraph object", {
  skip_if_not_installed("igraph")
  res <- cooccurrence(.test_list, output = "igraph")
  expect_true(igraph::is_igraph(res))
})

test_that("output = 'cograph' returns cograph_network", {
  skip_if_not_installed("cograph")
  res <- cooccurrence(.test_list, output = "cograph")
  expect_true(inherits(res, "cograph_network"))
})

# ========================================
# 11. Counting methods
# ========================================

# Transactions: {A,B,C} (n=3), {B,C} (n=2), {A,C} (n=2)

test_that("counting = 'full' gives raw counts (default)", {
  res <- cooccurrence(.test_list, counting = "full")
  ab <- res[res$from == "A" & res$to == "B", ]
  expect_equal(ab$weight, 1)
  expect_equal(ab$count, 1L)
})

test_that("counting = 'fractional' weights by 1/(n-1)", {
  # {A,B,C}: n=3, each pair gets 1/(3-1) = 0.5
  # {B,C}: n=2, each pair gets 1/(2-1) = 1
  # {A,C}: n=2, each pair gets 1
  # A-B: only in {A,B,C} → 0.5
  # B-C: in {A,B,C}(0.5) + {B,C}(1.0) = 1.5
  # A-C: in {A,B,C}(0.5) + {A,C}(1.0) = 1.5
  res <- cooccurrence(.test_list, counting = "fractional")
  ab <- res[res$from == "A" & res$to == "B", ]
  bc <- res[res$from == "B" & res$to == "C", ]
  ac <- res[res$from == "A" & res$to == "C", ]
  expect_equal(ab$weight, 0.5, tolerance = 1e-10)
  expect_equal(bc$weight, 1.5, tolerance = 1e-10)
  expect_equal(ac$weight, 1.5, tolerance = 1e-10)
  # count is always raw
  expect_equal(ab$count, 1L)
  expect_equal(bc$count, 2L)
})

test_that("counting = 'fractional' + similarity works", {
  res <- cooccurrence(.test_list, counting = "fractional", similarity = "jaccard")
  expect_true(nrow(res) > 0)
  # Weights should differ from full counting
  res_full <- cooccurrence(.test_list, counting = "full", similarity = "jaccard")
  expect_false(isTRUE(all.equal(res$weight, res_full$weight)))
})

test_that("counting with long/bipartite format", {
  df <- data.frame(
    doc = c(1, 1, 1, 2, 2, 3, 3),
    item = c("A", "B", "C", "B", "C", "A", "C"),
    stringsAsFactors = FALSE
  )
  res <- cooccurrence(df, field = "item", by = "doc", counting = "fractional")
  ab <- res[res$from == "A" & res$to == "B", ]
  expect_equal(ab$weight, 0.5, tolerance = 1e-10)
})

# ========================================
# 11b. field without sep (field_only format)
# ========================================

test_that("field without sep treats each column value as one item", {
  df <- data.frame(g1 = c("A", "B", "A"), g2 = c("B", "C", "C"),
                   stringsAsFactors = FALSE)
  expect_warning(res <- cooccurrence(df, field = c("g1", "g2")),
                 "without `sep`")
  expect_s3_class(res, "cooccurrence")
  expect_equal(nrow(res), 3L)
  expect_true(all(c("A", "B", "C") %in% c(res$from, res$to)))
})

test_that("field without sep ignores other columns", {
  df <- data.frame(id = 1:3, x = c("A", "B", "A"), y = c("B", "C", "C"),
                   stringsAsFactors = FALSE)
  expect_warning(res <- cooccurrence(df, field = c("x", "y")),
                 "without `sep`")
  items <- unique(c(res$from, res$to))
  expect_false(any(items %in% c("1", "2", "3")))
})

test_that("single field without sep gives one-item transactions (no edges)", {
  df <- data.frame(genre = c("Action", "Comedy", "Action"),
                   stringsAsFactors = FALSE)
  expect_warning(res <- cooccurrence(df, field = "genre"), "without `sep`")
  expect_equal(nrow(res), 0L)
})

test_that("field without sep matches multi_delimited with no delimiters", {
  df <- data.frame(g1 = c("A", "B", "A"), g2 = c("B", "C", "C"),
                   stringsAsFactors = FALSE)
  expect_warning(res_field <- cooccurrence(df, field = c("g1", "g2")),
                 "without `sep`")
  res_multi <- cooccurrence(df, field = c("g1", "g2"), sep = ";")
  key <- function(r) paste(pmin(r$from, r$to), pmax(r$from, r$to))
  r1 <- res_field[order(key(res_field)), ]
  r2 <- res_multi[order(key(res_multi)), ]
  expect_equal(r1$count, r2$count)
  expect_equal(r1$weight, r2$weight)
})

test_that("field without sep works with similarity", {
  df <- data.frame(g1 = c("A", "B", "A"), g2 = c("B", "C", "C"),
                   stringsAsFactors = FALSE)
  expect_warning(
    res <- cooccurrence(df, field = c("g1", "g2"), similarity = "jaccard"),
    "without `sep`"
  )
  expect_true(nrow(res) > 0)
  expect_true(all(res$weight >= 0 & res$weight <= 1))
})

test_that("field without sep handles NAs", {
  df <- data.frame(g1 = c("A", "B", NA), g2 = c("B", NA, "C"),
                   stringsAsFactors = FALSE)
  expect_warning(res <- cooccurrence(df, field = c("g1", "g2")),
                 "without `sep`")
  expect_s3_class(res, "cooccurrence")
})

test_that("field without sep suggests separator when found", {
  df <- data.frame(items = c("A;B;C", "B;C", "A;C"),
                   stringsAsFactors = FALSE)
  expect_warning(cooccurrence(df, field = "items"),
                 'sep = ";"')
})

test_that("field without sep suggests comma separator", {
  df <- data.frame(items = c("A,B,C", "B,C", "A,C"),
                   stringsAsFactors = FALSE)
  expect_warning(cooccurrence(df, field = "items"),
                 'sep = ","')
})

test_that("field without sep gives generic warning when no separator found", {
  df <- data.frame(g1 = c("A", "B"), g2 = c("C", "D"),
                   stringsAsFactors = FALSE)
  expect_warning(cooccurrence(df, field = c("g1", "g2")),
                 "without `sep`")
})

# ========================================
# 12. print.cooccurrence
# ========================================

test_that("print.cooccurrence shows header and edges", {
  res <- cooccurrence(.test_list)
  out <- capture.output(print(res))
  expect_true(any(grepl("cooccurrence:", out)))
  expect_true(any(grepl("nodes", out)))
  expect_true(any(grepl("edges", out)))
  expect_true(any(grepl("transactions", out)))
})

test_that("print.cooccurrence returns x invisibly", {
  res <- cooccurrence(.test_list)
  out <- withVisible(capture.output(ret <- print(res)))
  expect_identical(ret, res)
})

test_that("print.cooccurrence shows similarity and scale", {
  res <- cooccurrence(.test_list, similarity = "jaccard", scale = "log")
  out <- capture.output(print(res))
  expect_true(any(grepl("similarity: jaccard", out)))
  expect_true(any(grepl("scale: log", out)))
})

test_that("print.cooccurrence with split_by shows group info", {
  df <- data.frame(
    grp = c("X", "X", "Y", "Y"),
    kw = c("A; B", "A; B", "C; D", "C; D"),
    stringsAsFactors = FALSE
  )
  res <- cooccurrence(df, field = "kw", sep = ";", split_by = "grp")
  out <- capture.output(print(res))
  expect_true(any(grepl("split_by: grp", out)))
  expect_true(any(grepl("2 groups", out)))
})

# ========================================
# 13. weight_by (weighted long format)
# ========================================

.theta <- data.frame(
  doc   = c("d1","d1","d1","d2","d2","d3","d3"),
  topic = c("T1","T2","T3","T1","T3","T2","T3"),
  prob  = c(0.6, 0.3, 0.1, 0.4, 0.6, 0.5, 0.5),
  stringsAsFactors = FALSE
)

test_that("weight_by returns cooccurrence object with correct columns", {
  res <- cooccurrence(.theta, field = "topic", by = "doc", weight_by = "prob")
  expect_s3_class(res, "cooccurrence")
  expect_true(all(c("from", "to", "weight", "count") %in% names(res)))
})

test_that("weight_by computes correct weighted co-occurrence values", {
  res <- cooccurrence(.theta, field = "topic", by = "doc", weight_by = "prob")
  get_w <- function(a, b) {
    row <- res[res$from == a & res$to == b | res$from == b & res$to == a, ]
    row$weight
  }
  expect_equal(get_w("T1", "T3"), 0.6*0.1 + 0.4*0.6, tolerance = 1e-10)
  expect_equal(get_w("T2", "T3"), 0.3*0.1 + 0.5*0.5, tolerance = 1e-10)
  expect_equal(get_w("T1", "T2"), 0.6*0.3,            tolerance = 1e-10)
})

test_that("weight_by count column reflects binary co-occurrence", {
  res <- cooccurrence(.theta, field = "topic", by = "doc", weight_by = "prob")
  # T1-T2 appear together only in d1 -> count = 1
  row <- res[res$from == "T1" & res$to == "T2" |
             res$from == "T2" & res$to == "T1", ]
  expect_equal(row$count, 1L)
  # T1-T3 appear in d1 and d2 -> count = 2
  row <- res[res$from == "T1" & res$to == "T3" |
             res$from == "T3" & res$to == "T1", ]
  expect_equal(row$count, 2L)
})

test_that("weight_by works with similarity normalization", {
  res <- cooccurrence(.theta, field = "topic", by = "doc",
                      weight_by = "prob", similarity = "cosine")
  expect_s3_class(res, "cooccurrence")
  expect_true(all(res$weight >= 0 & res$weight <= 1))
})

test_that("weight_by respects min_occur", {
  res <- cooccurrence(.theta, field = "topic", by = "doc",
                      weight_by = "prob", min_occur = 2)
  # T1-T2 pair: T2 appears in d1 and d3 (2 docs), T1 in d1 and d2 (2 docs)
  # All topics appear in >= 2 docs, so all 3 edges remain
  expect_equal(nrow(res), 3L)
})

test_that("weight_by errors on non-long format", {
  df <- data.frame(items = c("A;B", "B;C"), w = c(1, 2),
                   stringsAsFactors = FALSE)
  expect_error(
    cooccurrence(df, field = "items", sep = ";", weight_by = "w"),
    "long format"
  )
})

test_that("weight_by works with split_by", {
  df <- data.frame(
    doc   = c("d1","d1","d2","d2","d3","d3"),
    topic = c("T1","T2","T1","T2","T1","T2"),
    prob  = c(0.6, 0.4, 0.3, 0.7, 0.5, 0.5),
    year  = c(2020, 2020, 2020, 2020, 2021, 2021),
    stringsAsFactors = FALSE
  )
  res <- cooccurrence(df, field = "topic", by = "doc",
                      weight_by = "prob", split_by = "year")
  expect_true("group" %in% names(res))
  expect_equal(sort(unique(res$group)), c("2020", "2021"))
})

test_that("print.cooccurrence respects n parameter", {
  res <- cooccurrence(.test_list)
  out <- capture.output(print(res, n = 1L))
  expect_true(any(grepl("more edges", out)))
})

test_that("print.cooccurrence handles zero edges", {
  res <- cooccurrence(list(c("A"), c("B"), c("C")))
  out <- capture.output(print(res))
  expect_true(any(grepl("no edges", out)))
})

test_that("print.cooccurrence omits similarity when 'none'", {
  res <- cooccurrence(.test_list, similarity = "none")
  out <- capture.output(print(res))
  expect_false(any(grepl("similarity:", out)))
})

# ========================================
# 13. summary.cooccurrence
# ========================================

test_that("summary.cooccurrence shows network stats", {
  res <- cooccurrence(.test_list, similarity = "jaccard")
  out <- capture.output(summary(res))
  expect_true(any(grepl("cooccurrence network", out)))
  expect_true(any(grepl("Nodes", out)))
  expect_true(any(grepl("Edges", out)))
  expect_true(any(grepl("Density", out)))
  expect_true(any(grepl("Transactions", out)))
  expect_true(any(grepl("Similarity.*jaccard", out)))
  expect_true(any(grepl("Weight range", out)))
  expect_true(any(grepl("Count range", out)))
  expect_true(any(grepl("Top nodes", out)))
})

test_that("summary.cooccurrence returns object invisibly", {
  res <- cooccurrence(.test_list)
  capture.output(ret <- summary(res))
  expect_identical(ret, res)
})

test_that("summary.cooccurrence shows scale when not 'none'", {
  res <- cooccurrence(.test_list, scale = "minmax")
  out <- capture.output(summary(res))
  expect_true(any(grepl("Scale.*minmax", out)))
})

test_that("summary.cooccurrence omits scale when 'none'", {
  res <- cooccurrence(.test_list)
  out <- capture.output(summary(res))
  expect_false(any(grepl("Scale", out)))
})

test_that("summary.cooccurrence handles zero edges", {
  res <- cooccurrence(list(c("A"), c("B"), c("C")))
  out <- capture.output(summary(res))
  expect_true(any(grepl("Edges.*: 0", out)))
  # Should NOT print weight/count/top nodes sections
  expect_false(any(grepl("Weight range", out)))
})

# ========================================
# 14. plot.cooccurrence
# ========================================

test_that("plot.cooccurrence heatmap works", {
  pdf(NULL); on.exit(dev.off(), add = TRUE)
  res <- cooccurrence(.test_list)
  expect_invisible(plot(res, type = "heatmap"))
})

test_that("plot.cooccurrence heatmap works when matrix attr is NULL", {
  pdf(NULL); on.exit(dev.off(), add = TRUE)
  res <- cooccurrence(.test_list)
  attr(res, "matrix") <- NULL
  expect_invisible(plot(res, type = "heatmap"))
})

test_that("plot.cooccurrence network requires igraph", {
  skip_if_not_installed("igraph")
  pdf(NULL); on.exit(dev.off(), add = TRUE)
  res <- cooccurrence(.test_list)
  expect_invisible(plot(res, type = "network"))
})

test_that("plot.cooccurrence returns x invisibly", {
  pdf(NULL); on.exit(dev.off(), add = TRUE)
  res <- cooccurrence(.test_list)
  ret <- plot(res, type = "heatmap")
  expect_identical(ret, res)
})

# ========================================
# 15. Converter edge cases
# ========================================

test_that("as_matrix rebuilds from edges when attribute is lost", {
  res <- cooccurrence(.test_list, similarity = "jaccard")
  attr(res, "matrix") <- NULL
  mat <- as_matrix(res)
  expect_true(is.matrix(mat))
  expect_equal(nrow(mat), 3L)
  expect_true(isSymmetric(mat))
  # Should match the weight values
  ab <- res[res$from == "A" & res$to == "B", ]
  expect_equal(mat["A", "B"], ab$weight)
})

test_that("as_matrix rebuilds raw from edges when attribute is lost", {
  res <- cooccurrence(.test_list, similarity = "jaccard")
  attr(res, "raw_matrix") <- NULL
  raw <- as_matrix(res, type = "raw")
  expect_true(is.matrix(raw))
  expect_equal(raw["A", "B"], 1)
  expect_equal(raw["A", "C"], 2)
})

test_that("as_igraph works when items attribute is NULL", {
  skip_if_not_installed("igraph")
  res <- cooccurrence(.test_list)
  attr(res, "items") <- NULL
  g <- as_igraph(res)
  expect_true(igraph::is_igraph(g))
  expect_equal(igraph::vcount(g), 3L)
})

test_that("as_cograph structure is complete", {
  skip_if_not_installed("cograph")
  res <- cooccurrence(.test_list)
  net <- as_cograph(res)
  expect_false(net$directed)
  expect_equal(net$n_edges, 3L)
  expect_true(is.matrix(net$weights))
  expect_true(all(c("id", "label", "name", "x", "y") %in% names(net$nodes)))
  expect_true(all(c("from", "to", "weight") %in% names(net$edges)))
  expect_equal(net$meta$source, "cooccure")
})

test_that("as_netobject structure is complete", {
  skip_if_not_installed("Nestimate")
  res <- cooccurrence(.test_list, similarity = "jaccard", threshold = 0)
  net <- as_netobject(res)
  expect_equal(net$method, "cooccurrence")
  expect_equal(net$params$similarity, "jaccard")
  expect_true(is.matrix(net$weights))
  expect_true(all(c("id", "label", "name") %in% names(net$nodes)))
  expect_true(all(c("from", "to", "weight") %in% names(net$edges)))
  expect_equal(net$meta$tna$method, "cooccurrence")
})

# ========================================
# 16. Converter error paths (missing packages)
# ========================================

test_that("as_tidygraph errors when tidygraph is not available", {
  skip_if(requireNamespace("tidygraph", quietly = TRUE),
          "tidygraph is installed; cannot test missing-package path")
  res <- cooccurrence(.test_list)
  expect_error(as_tidygraph(res), "tidygraph")
})

test_that("as_netobject errors when Nestimate is not available", {
  skip_if(requireNamespace("Nestimate", quietly = TRUE),
          "Nestimate is installed; cannot test missing-package path")
  res <- cooccurrence(.test_list)
  expect_error(as_netobject(res), "Nestimate")
})

# ========================================
# 17. as_cograph / as_netobject with zero edges
# ========================================

test_that("as_cograph handles zero edges", {
  skip_if_not_installed("cograph")
  res <- cooccurrence(list(c("A"), c("B")))
  net <- as_cograph(res)
  expect_equal(net$n_edges, 0L)
  expect_equal(nrow(net$edges), 0L)
  expect_true(all(c("from", "to", "weight") %in% names(net$edges)))
})

test_that("as_netobject handles zero edges", {
  skip_if_not_installed("Nestimate")
  res <- cooccurrence(list(c("A"), c("B")))
  net <- as_netobject(res)
  expect_equal(net$n_edges, 0L)
  expect_equal(nrow(net$edges), 0L)
})

# ========================================
# 18. as_tidygraph full path
# ========================================

test_that("as_tidygraph produces correct tbl_graph", {
  skip_if_not_installed("tidygraph")
  skip_if_not_installed("igraph")
  res <- cooccurrence(.test_list)
  tg <- as_tidygraph(res)
  expect_true(inherits(tg, "tbl_graph"))
  expect_equal(igraph::vcount(tg), 3L)
  expect_equal(igraph::ecount(tg), 3L)
})

# ========================================
# 19. as_igraph edge attributes & vertex details
# ========================================

test_that("as_igraph preserves edge weights and counts", {
  skip_if_not_installed("igraph")
  res <- cooccurrence(.test_list, similarity = "jaccard")
  g <- as_igraph(res)
  expect_true("weight" %in% igraph::edge_attr_names(g))
  expect_true("count" %in% igraph::edge_attr_names(g))
  expect_equal(igraph::ecount(g), nrow(res))
})

test_that("as_igraph includes isolated nodes from items attr", {
  skip_if_not_installed("igraph")
  # min_occur filters D from edges, but items attr includes all original items
  trans <- list(c("A", "B"), c("A", "B"), c("A", "D"))
  res <- cooccurrence(trans, min_occur = 2L)
  # Items attr still has D if it appeared in transactions
  g <- as_igraph(res)
  expect_true(igraph::is_igraph(g))
})

# ========================================
# 20. as_cograph detailed checks
# ========================================

test_that("as_cograph nodes have sequential ids", {
  skip_if_not_installed("cograph")
  res <- cooccurrence(.test_list)
  net <- as_cograph(res)
  expect_equal(net$nodes$id, seq_len(net$n_nodes))
})

test_that("as_cograph edges reference valid node ids", {
  skip_if_not_installed("cograph")
  res <- cooccurrence(.test_list)
  net <- as_cograph(res)
  all_ids <- net$nodes$id
  expect_true(all(net$edges$from %in% all_ids))
  expect_true(all(net$edges$to %in% all_ids))
})

test_that("as_cograph weights matrix matches normalized matrix", {
  skip_if_not_installed("cograph")
  res <- cooccurrence(.test_list, similarity = "cosine")
  net <- as_cograph(res)
  mat <- as_matrix(res, type = "normalized")
  expect_equal(net$weights, mat)
})

# ========================================
# 21. as_matrix thorough
# ========================================

test_that("as_matrix diagonal is zero (no self-loops)", {
  res <- cooccurrence(.test_list)
  mat <- as_matrix(res)
  expect_true(all(diag(mat) == 0))
  raw <- as_matrix(res, type = "raw")
  expect_true(all(diag(raw) == 0))
})

test_that("as_matrix dimnames are sorted unique items", {
  res <- cooccurrence(.test_list)
  mat <- as_matrix(res)
  items <- sort(unique(c(res$from, res$to)))
  expect_equal(rownames(mat), items)
  expect_equal(colnames(mat), items)
})

test_that("as_matrix rebuilt normalized matches stored matrix", {
  res <- cooccurrence(.test_list, similarity = "jaccard")
  ## Stored attribute is a sparse Matrix in the current engine; densify for
  ## comparison against the dense fallback rebuilt from the edge list.
  stored <- as.matrix(attr(res, "matrix"))
  attr(res, "matrix") <- NULL
  rebuilt <- as_matrix(res)
  # Diagonal is lost when rebuilding from edges (no self-edges), so compare off-diagonal
  diag(stored) <- 0
  expect_equal(rebuilt, stored)
})

test_that("as_matrix rebuilt raw is symmetric", {
  res <- cooccurrence(.test_list, similarity = "cosine")
  attr(res, "raw_matrix") <- NULL
  raw <- as_matrix(res, type = "raw")
  expect_true(isSymmetric(raw))
})

test_that("as_matrix with scaled result returns scaled values", {
  res <- cooccurrence(.test_list, similarity = "jaccard", scale = "minmax")
  mat <- as_matrix(res)
  expect_true(all(mat >= 0))
  expect_true(all(mat <= 1))
})

test_that("as_matrix on zero-edge result returns zero off-diagonal", {
  res <- cooccurrence(list(c("A"), c("B"), c("C")))
  mat <- as_matrix(res, type = "raw")
  diag(mat) <- 0
  expect_true(all(mat == 0))
})

test_that("as_matrix on split_by falls back to rebuild from edges", {
  df <- data.frame(
    grp = c("X", "X", "Y", "Y"),
    kw = c("A; B", "A; B", "C; D", "C; D"),
    stringsAsFactors = FALSE
  )
  res <- cooccurrence(df, field = "kw", sep = ";", split_by = "grp")
  # split_by does not store per-group matrices in attr, but the combined
  # result may not have a matrix attr → rebuild path
  mat <- as_matrix(res)
  expect_true(is.matrix(mat))
  expect_true(isSymmetric(mat))
})

# ========================================
# 22. Generic dispatch errors
# ========================================

test_that("as_matrix errors on non-cooccurrence input", {
  expect_error(as_matrix(42))
})

test_that("as_igraph errors on non-cooccurrence input", {
  expect_error(as_igraph(42))
})

test_that("as_cograph errors on non-cooccurrence input", {
  expect_error(as_cograph(42))
})

test_that("as_tidygraph errors on non-cooccurrence input", {
  expect_error(as_tidygraph(42))
})

test_that("as_netobject errors on non-cooccurrence input", {
  expect_error(as_netobject(42))
})

# ========================================
# 23. as_igraph thorough
# ========================================

test_that("as_igraph vertex names match items attribute", {
  skip_if_not_installed("igraph")
  res <- cooccurrence(.test_list)
  g <- as_igraph(res)
  expect_setequal(igraph::V(g)$name, attr(res, "items"))
})

test_that("as_igraph edge weights match cooccurrence weights", {
  skip_if_not_installed("igraph")
  res <- cooccurrence(.test_list, similarity = "jaccard")
  g <- as_igraph(res)
  el <- igraph::as_data_frame(g, what = "edges")
  # Sort both for comparison
  el <- el[order(el$from, el$to), ]
  co_sorted <- res[order(res$from, res$to), ]
  expect_equal(el$weight, co_sorted$weight)
  expect_equal(el$count, co_sorted$count)
})

test_that("as_igraph with zero edges gives edgeless graph", {
  skip_if_not_installed("igraph")
  res <- cooccurrence(list(c("A"), c("B"), c("C")))
  g <- as_igraph(res)
  expect_equal(igraph::ecount(g), 0L)
  # Vertices still present if items attr exists
  items <- attr(res, "items")
  if (!is.null(items)) {
    expect_true(igraph::vcount(g) >= length(items))
  }
})

test_that("as_igraph with similarity + scale produces valid graph", {
  skip_if_not_installed("igraph")
  res <- cooccurrence(.test_list, similarity = "dice", scale = "sqrt")
  g <- as_igraph(res)
  expect_true(igraph::is_igraph(g))
  expect_true(all(igraph::E(g)$weight >= 0))
})

test_that("as_igraph roundtrip preserves edge count", {
  skip_if_not_installed("igraph")
  res <- cooccurrence(.test_list)
  g <- as_igraph(res)
  el <- igraph::as_data_frame(g, what = "edges")
  expect_equal(nrow(el), nrow(res))
})

# ========================================
# 24. as_cograph thorough
# ========================================

test_that("as_cograph node labels match sorted items", {
  skip_if_not_installed("cograph")
  res <- cooccurrence(.test_list)
  net <- as_cograph(res)
  mat <- as_matrix(res, type = "normalized")
  expect_equal(net$nodes$label, colnames(mat))
  expect_equal(net$nodes$name, colnames(mat))
})

test_that("as_cograph edge count matches non-zero upper triangle", {
  skip_if_not_installed("cograph")
  res <- cooccurrence(.test_list, similarity = "jaccard")
  net <- as_cograph(res)
  mat <- as_matrix(res, type = "normalized")
  expected_edges <- sum(upper.tri(mat) & mat != 0)
  expect_equal(net$n_edges, expected_edges)
  expect_equal(nrow(net$edges), expected_edges)
})

test_that("as_cograph meta$layout is NULL", {
  skip_if_not_installed("cograph")
  res <- cooccurrence(.test_list)
  net <- as_cograph(res)
  expect_null(net$meta$layout)
})

test_that("as_cograph with single pair", {
  skip_if_not_installed("cograph")
  res <- cooccurrence(list(c("X", "Y"), c("X", "Y")))
  net <- as_cograph(res)
  expect_equal(net$n_nodes, 2L)
  expect_equal(net$n_edges, 1L)
  expect_equal(nrow(net$nodes), 2L)
  expect_equal(nrow(net$edges), 1L)
})

test_that("as_cograph with different similarity measures", {
  skip_if_not_installed("cograph")
  for (sim in c("none", "jaccard", "cosine", "dice")) {
    res <- cooccurrence(.test_list, similarity = sim)
    net <- as_cograph(res)
    expect_true(inherits(net, "cograph_network"))
    expect_equal(net$n_nodes, 3L)
  }
})

test_that("as_cograph node x and y are NA", {
  skip_if_not_installed("cograph")
  res <- cooccurrence(.test_list)
  net <- as_cograph(res)
  expect_true(all(is.na(net$nodes$x)))
  expect_true(all(is.na(net$nodes$y)))
})

test_that("as_cograph edges$from and edges$to are integer", {
  skip_if_not_installed("cograph")
  res <- cooccurrence(.test_list)
  net <- as_cograph(res)
  expect_type(net$edges$from, "integer")
  expect_type(net$edges$to, "integer")
})

# ========================================
# 14. Large-dataset scalability
# ========================================

test_that("cooccurrence scales to wide sparse inputs without densifying", {
  ## Regression guard: the engine must stay in sparse representation.
  ## The dense predecessor allocated n * k logicals plus a k * k weight matrix;
  ## at these sizes that would be ~60 GB and fail with `vector memory limit`.
  skip_on_cran()
  n_docs  <- 20000L
  k_items <- 20000L
  items_per_doc <- 20L  # typical citation count

  set.seed(42)
  transactions <- lapply(seq_len(n_docs), function(i) {
    paste0("R", sample.int(k_items, items_per_doc))
  })

  ## Run both counting methods; just assert we return an edge list without
  ## blowing up memory or time.
  t0 <- Sys.time()
  res_full <- cooccurrence(transactions, counting = "full")
  res_frac <- cooccurrence(transactions, counting = "fractional")
  elapsed <- as.numeric(difftime(Sys.time(), t0, units = "secs"))

  expect_s3_class(res_full, "cooccurrence")
  expect_s3_class(res_frac, "cooccurrence")
  expect_gt(nrow(res_full), 0L)
  expect_gt(nrow(res_frac), 0L)

  ## The stored matrix attribute must be sparse, not a dense k x k base matrix.
  expect_true(inherits(attr(res_full, "matrix"), "Matrix"))
  expect_true(inherits(attr(res_frac, "matrix"), "Matrix"))

  ## Sanity ceiling on wall time — catches accidental dense regressions that
  ## would take minutes instead of seconds.
  expect_lt(elapsed, 120)
})

---
title: "Co-occurrence Networks with IMDB Movie Data"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Co-occurrence Networks with IMDB Movie Data}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include=FALSE}
knitr::opts_chunk$set(
  collapse = TRUE, comment = "#>", fig.width = 8, fig.height = 6,
  message = FALSE, warning = FALSE
)
```

This tutorial demonstrates every feature of the `cooccure` package using 1,000 highly-rated IMDB movies (rating $\geq$ 7.0, $\geq$ 1,000 votes, 1970--2024).

Starting from raw tabular data, we build genre co-occurrence networks, construct actor collaboration networks, compare co-occurrence patterns across decades and rating bands, apply different similarity measures and counting methods, and export results to `Gephi`, `igraph`, and `cograph` for visualization and downstream analysis.

## Data

```{r load-data}
library(cooccure)
```

The dataset contains the title of the movie, the year and decade, the `genres` (comma-separated), the average rating, the number of votes, and the rating band.
```{r}
head(movies)
```


## 1. Genre co-occurrence (delimited field)

Each movie has a comma-delimited `genres` column, where genres that appear in the same movie are connected. To analyze their co-occurrence  we need to call `cooccurrence` and specify the `field` argument specifying the column and `sep` specifying the delimiter that separates values within that field.

```{r genre-basic}
cooccurrence(movies, field = "genres", sep = ",")
```

The result shows 22 genre nodes and 129 edges. *Drama* dominates the top pairs simply because it is the most frequent genre, inflating co-occurrence with nearly every other genre regardless of actual association strength. Jaccard similarity normalizes for this, so the strongest edges reflect true affinity rather than frequency. A directed co-occurrence graph is then built and its degree distribution plotted as a bar chart, showing how many nodes share each degree value and whether connectivity is concentrated or evenly spread across genres.

```{r genre-jaccard}
library(cograph)

Net <- co(movies, field = "genres", sep = ",", similarity = "jaccard")

Gr<- as_cograph(Net, directed = TRUE)

degree_distribution(Gr)
```

### Comparing similarity measures

Each similarity measure surfaces different structure in the data, prioritizing different aspects of co-occurrence. The examples below show the top 3 pairs under each measure using a movie genres dataset.


 - `similarity = "none"` — Raw counts favor the most frequent genre pairs. *Drama* dominates because it is the most common genre, so its co-occurrences with *Comedy* and *Romance* rank highest regardless of how strongly the genres are actually associated.

```{r sim-none}
co(movies, field = "genres", sep = ",", similarity = "none", top_n = 3)
```

 - `similarity = "jaccard"` — Normalizing by the union of occurrences brings less frequent but more tightly associated pairs to the top. *Adventure*–*Animation* emerges as the strongest pair, suggesting these genres appear together more consistently relative to how often either appears alone.

```{r sim-jaccard}
co(movies, field = "genres", sep = ",", similarity = "jaccard", top_n = 3)
```

 - `similarity = "cosine"` — Similar to Jaccard but less strict, cosine also elevates *Adventure*–*Animation* while keeping *Drama*–*Romance* in the top 3, reflecting its more lenient treatment of frequency differences between genres.

```{r sim-cosine}
co(movies, field = "genres", sep = ",", similarity = "cosine", top_n = 3)
```
 
 - `similarity = "inclusion"` — Dividing by the less common genre's frequency surfaces subset relationships. *Documentary*–*News* and *History*–*News* both score 1, meaning *News* movies are always also tagged as *Documentary* or *History*—*News* is a niche genre that only appears alongside broader ones.

```{r sim-inclusion}
co(movies, field = "genres", sep = ",", similarity = "inclusion", top_n = 3)
```

 - `similarity = "association"` — Discounting by the product of individual frequencies reveals pairs that co-occur far more than chance would predict. Rare genre combinations like *History*–*News* and *Documentary*–*News* top the list because their co-occurrence is disproportionately high relative to their individual frequencies.

```{r sim-association}
co(movies, field = "genres", sep = ",", similarity = "association", top_n = 3)
```

 - `similarity = "dice"` — Results closely mirror Jaccard, with the same top pairs appearing in the same order but with slightly higher weights — reflecting Dice's more lenient arithmetic mean normalization. *Adventure*-*Animation*, *Action*-*Crime*, and *Comedy*-*Drama* remain the strongest genre affinities, with *Comedy*-*Drama* recording the highest raw co-occurrence count at 159 despite its comparatively lower weight, indicating that both genres are individually common but not exclusively paired.

```{r sim-dice}
co(movies, field = "genres", sep = ",", similarity = "dice", top_n = 3)
```

 - `similarity = "equivalence"` — Squaring the cosine amplifies differences between strong and weak pairs, pushing weaker associations further down the rankings. *Adventure*-*Animation* retains the top spot, but *Drama*–*Romance* displaces *Action*-*Crime* into third place, suggesting that while *Action*-*Crime* has a higher raw count, *Drama*–*Romance* represents a tighter and more exclusive pairing once the penalty for common genres is compounded.

```{r sim-equivalence}
co(movies, field = "genres", sep = ",", similarity = "equivalence", top_n = 3)
```

### Which similarity to use?

- **Exploratory work**: Start with `"none"` to see raw counts and understand the data, then try `"jaccard"` or `"cosine"` for a balanced view.
- **Bibliometric and scientometric networks**: `"association"` is recommended by van Eck & Waltman (2009) because it correctly accounts for the expected number of co-occurrences under independence. Two items that are both very frequent will naturally co-occur often; association strength discounts this, revealing which pairs co-occur *more than chance alone would predict*.
- **Detecting hierarchical or subset structure**: `"inclusion"` (the Simpson coefficient) reveals when one item almost always appears with another --- useful for identifying items that are subsets of broader categories, or for detecting dependency relationships.
- **Binary presence/absence networks**: `"jaccard"` or `"dice"` when you only care *whether* items co-occur, not *how often*. Jaccard is stricter (penalizes unbalanced pairs more); Dice is more lenient.
- **Scale-invariant comparison**: `"cosine"` is invariant to absolute frequency --- useful when comparing co-occurrence patterns across datasets of different sizes.
- **Strict filtering**: `"equivalence"` (cosine squared) amplifies differences --- pairs with weak overlap get pushed closer to zero, retaining only the strongest associations.

## 2. Counting methods

By default, each transaction contributes equally to co-occurrence counts: if a movie has genres A, B, and C, each pair adds 1. This means movies with many genres inflate the network — a 5-genre movie creates 10 pairs, while a 2-genre movie creates only 1, giving multi-genre movies disproportionate influence.

**Fractional counting** addresses this by weighting each pair by $1/(n-1)$, where $n$ is the number of items in the transaction, so every transaction contributes equally regardless of how many genres it contains.

```{r counting-full}
co(movies, field = "genres", sep = ",", top_n = 5)
```

```{r counting-fractional}
co(movies, field = "genres", sep = ",", counting = "fractional", top_n = 5)
```

The top pairs remain the same under both methods, but the weights are lower under fractional counting. For example, *Comedy*-*Drama* drops from 159 to 107.5, reflecting the downweighting of multi-genre movies that contributed to this pair. Fractional counting is particularly important when some transactions contain many items while others contain few, as it prevents high-cardinality transactions from dominating the network.

## 3. Scaling

Scaling compresses or transforms the weight distribution after similarity normalization, making it easier to visualize or use in downstream analysis. Scaling can be applied on its own or combined with any similarity measure.

 - `scale = "log"` — Applies a natural log transformation, compressing the heavy tail of the distribution. The ranking of pairs is preserved but the gap between frequent and infrequent pairs is reduced. *Comedy*-*Drama* and Drama*-*Romance* remain at the top, but their weights are now much closer together than raw counts would suggest.

```{r scale-log}
co(movies, field = "genres", sep = ",", scale = "log", top_n = 5)
```

 - `scale = "minmax"` — Rescales all weights to the range [0, 1], where the strongest pair scores 1 and all others are expressed relative to it. Useful for comparing networks of different sizes or when absolute counts are not meaningful. *Adventure*-*Animation* scores a perfect 1 as the strongest Jaccard pair in the network. *Biography*-*Documentary* reaches the top 5 with a raw count of only 44, indicating that when these genres appear together they do so with high exclusivity relative to how often either appears alone.

```{r scale-minmax}
co(movies, field = "genres", sep = ",", similarity = "jaccard", scale = "minmax", top_n = 5)
```

 - `scale = "binary"` — Converts all positive weights to 1, producing a presence/absence network. The top pairs are no longer ranked by strength but simply by whether they co-occur at all. *Action*-*Adventure* dominate simply because they appear together across a wide range of combinations.

```{r scale-binary}
co(movies, field = "genres", sep = ",", scale = "binary", top_n = 5)
```

 - `scale = "sqrt"` — Applies a square root transformation, providing a milder compression than log. The ranking is preserved and the distribution is slightly less skewed than the raw counts. *Comedy*-*Drama* leads with a weight of 12.6 compared to *Comedy*-*Romance* at 7.9, a smaller proportional gap than the raw count difference of 159 versus 63 would imply.

```{r scale-sqrt}
co(movies, field = "genres", sep = ",", scale = "sqrt", top_n = 5)
```

Scaling can be combined with any similarity measure and with filtering arguments. The example below applies association strength followed by log scaling, retaining only genres appearing in at least 20 movies. *Drama*-heavy pairs drop out entirely once popularity is penalized, and niche but tightly linked pairs like *Adventure*-*Animation*, *History*-*War*, and *Mystery*-*Thriller* rise to the top, reflecting genres that genuinely cluster together rather than simply co-occurring by volume.

```{r scale-combined}
co(movies, field = "genres", sep = ",",
   similarity = "association", scale = "log", min_occur = 20, top_n = 5)
```

## 4. Filtering

Three filtering arguments control which edges appear in the result. They can be used independently or combined.

 - `min_occur` — Drops any genre appearing in fewer than the specified number of transactions before co-occurrences are computed, removing rare items that would otherwise inflate the edge count. Here only movies with co-occurrence of minimum 20 are kept. 

```{r min-occur}
co(movies, field = "genres", sep = ",", similarity = "jaccard", min_occur = 20)
```

 - `threshold` — Retains only edges with a weight at or above the specified value, applied after similarity normalization and scaling. Here only pairs with a Jaccard similarity above 0.15 are kept.

```{r threshold}
co(movies, field = "genres", sep = ",", similarity = "jaccard", threshold = 0.15)
```

 - `top_n` — Keeps only the n strongest edges by weight, regardless of their absolute value.

```{r top-n}
co(movies, field = "genres", sep = ",", similarity = "jaccard", top_n = 10)
```

All three thresholds can be combined for fine-grained control over network size and density:

```{r combined-filter}
co(movies, field = "genres", sep = ",",
   similarity = "association", counting = "fractional",
   min_occur = 15, threshold = 0.001, top_n = 20)
```

## 5. Actor co-occurrence (long/bipartite format)

Actors who appear in the same movie are connected. The data has one row per actor–movie pair, making it a natural fit for the long/bipartite format. The `field` argument specifies the entity column and `by` specifies the grouping column. 

```{r actor-net}
co(actors, field = "actor", by = "tconst",
   similarity = "jaccard", min_occur = 3, threshold = 0.1)
```

Julia Bache-Wiig and Robin Ottersen score 1.0, meaning they appear together in every movie either of them appears in — a perfect Jaccard score. Applying fractional counting downweights pairs from movies with large casts, though in this case the results are unchanged because the dataset is small and cast sizes are similar across movies.

```{r actor-frac}
co(actors, field = "actor", by = "tconst",
   similarity = "jaccard", counting = "fractional",
   min_occur = 3, threshold = 0.05)
```

## 6. Splitting by groups

The split_by argument computes a separate co-occurrence network for each level of a grouping variable and returns all results in a single data frame with a group column.

 - `split_by = "decade"` — Each decade gets its own Jaccard-weighted genre network. The dominant pairs shift across decades, reflecting how genre combinations have changed over time.

```{r split-decade}
co(movies, field = "genres", sep = ",",
   split_by = "decade", similarity = "jaccard",
   min_occur = 5, top_n = 5)
```

Individual groups can be extracted by filtering the group column:

```{r filter-decade}
decades <- co(movies, field = "genres", sep = ",",
              split_by = "decade", similarity = "jaccard",
              min_occur = 5, top_n = 5)

decades[decades$group == "2010s", ]

```

 - `split_by = "rating_band"` — Splitting by rating band reveals whether highly rated movies have different genre co-occurrence patterns than average-rated ones. *Documentary*–*Music* is the strongest pair among top-rated movies, while *Adventure*–*Animation* leads among the 7–7.9 band.

```{r split-rating}
movies$rating_band <- ifelse(movies$averageRating >= 8, "8+", "7-7.9")
co(movies, field = "genres", sep = ",",
   split_by = "rating_band", similarity = "jaccard",
   min_occur = 10, top_n = 5)
```

## 7. Output formats

The default output format returns a tidy data frame with `from`, `to`, `weight`, and `count` columns, ready for further analysis or visualization.

```{r out-default}
co(movies, field = "genres", sep = ",", top_n = 5)
```

### Gephi

The Gephi output format returns a data frame formatted for direct import into Gephi, with `Source`, `Target`, `Weight`, `Type`, and `Count` columns. The result can be written straight to CSV.

```{r out-gephi}
co(movies, field = "genres", sep = ",",
   similarity = "jaccard", output = "gephi", top_n = 10)
```

```{r gephi-export, eval=FALSE}
write.csv(
  co(movies, field = "genres", sep = ",", similarity = "jaccard", output = "gephi"),
  "genre_network.csv", row.names = FALSE
)
```

### cograph (with Gephi layout)

The cograph output format returns a `cograph_network` object that can be passed directly to `splot()` for visualization. The layout argument controls the node placement algorithm — `"fr"` uses Fruchterman-Reingold, and `scale_nodes_by = "degree"` sizes nodes by their degree centrality.

```{r out-cograph, fig.width=8, fig.height=8}
library(cograph)
net <- co(movies, field = "genres", sep = ",",
          similarity = "jaccard", min_occur = 20, output = "cograph")
splot(net, layout = "fr", scale_nodes_by = "degree")
```

Additional styling arguments control `edge_width_range`, `label_size`, `node_color`, and `node_border_width`:

```{r cograph-styled, fig.width=8, fig.height=8}
library(cograph)
splot(net, layout = "gephi", label_size = .8, label_fontface = "bold",
           node_fill = "#F9C22E",  
           node_border_width = 0.0001, edge_color = "black",
           scale_nodes_by = "degree", edge_width_range = c(0.1:4))
```

### igraph

The igraph output format returns an `igraph` object. All standard igraph functions work on the result without any conversion.

```{r out-igraph}
g <- co(movies, field = "genres", sep = ",",
        similarity = "jaccard", min_occur = 20, output = "igraph")
g
```
Centrality measures can be computed directly on the `igraph` object:

```{r igraph-metrics}
igraph::degree(g)
igraph::betweenness(g)
```

### Matrix

The matrix output format returns a square co-occurrence matrix where rows and columns are items and each cell contains the similarity weight between the corresponding pair. This is useful for downstream analysis that expects a matrix input, such as clustering or heatmap visualization.

```{r out-matrix}
mat <- co(movies, field = "genres", sep = ",",
          similarity = "jaccard", min_occur = 20, output = "matrix")
round(mat[1:6, 1:6], 3)
```

## 8. Converters

Converters transform a `cooccurrence` result into other formats after the fact, without re-running the computation. This is useful when you want to start with the default tidy data frame and convert to a specific format only when needed.

`as_matrix()` converts the result to a square similarity matrix, where each cell contains the Jaccard weight between the corresponding pair of genres:

```{r conv-matrix}
result <- co(movies, field = "genres", sep = ",",
             similarity = "jaccard", min_occur = 20)
as_matrix(result)
```
`as_matrix(result, type = "raw")` returns the raw co-occurrence count matrix instead, with no similarity normalization applied:

```{r conv-raw}
as_matrix(result, type = "raw")
```
`as_igraph()` converts the result to an igraph object, giving access to the full igraph ecosystem for further network analysis:

```{r conv-igraph}
as_igraph(result)
```

## 9. Six input formats, one result

The same data can be provided in different formats — `cooccure` auto-detects the format and produces identical results regardless of which representation is used. The four examples below all compute the same genre co-occurrence network from the same underlying data.

*Delimited field* is the most common format. A single column contains all genres as a comma-separated string, one row per movie.

```{r fmt-delimited}
res1 <- co(movies, field = "genres", sep = ",")
```

*Long/bipartite* format uses one row per genre–movie pair. The data is first reshaped from wide to long, then passed to `co()` with `field` specifying the genre column and by specifying the movie identifier.

```{r fmt-long}
genre_long <- do.call(rbind, lapply(seq_len(nrow(movies)), function(i) {
  gs <- trimws(strsplit(movies$genres[i], ",")[[1]])
  data.frame(movie_id = movies$tconst[i], genre = gs, stringsAsFactors = FALSE)
}))
res2 <- co(genre_long, field = "genre", by = "movie_id")
```

*Binary matrix* uses a document-term matrix where rows are movies, columns are genres, and values are 0 or 1. Auto-detected when all values are binary and no `field`, `by`, or `sep` arguments are provided.

```{r fmt-binary}
all_genres <- sort(unique(genre_long$genre))
bin <- matrix(0L, nrow = nrow(movies), ncol = length(all_genres),
              dimnames = list(movies$tconst, all_genres))
for (i in seq_len(nrow(genre_long))) {
  row <- match(genre_long$movie_id[i], movies$tconst)
  bin[row, genre_long$genre[i]] <- 1L
}
res3 <- co(bin)
```
*List of character vectors* is the most direct format, where each list element is a character vector of genres for one movie.

```{r fmt-list}
res4 <- co(lapply(strsplit(movies$genres, ","), trimws))
```

All four produce identical weights, confirming that the format choice is purely a matter of convenience:

```{r fmt-verify}
all.equal(res1$weight, res2$weight)
all.equal(res1$weight, res3$weight)
all.equal(res1$weight, res4$weight)
```

## 10. Complete pipeline

All steps can be combined in a single call and piped directly to `splot()` for visualization. The example below applies Jaccard similarity, fractional counting, min-max scaling, and filtering in one expression, converting the result to a `cograph_network` and rendering it without any intermediate objects.

```{r pipeline, fig.width=8, fig.height=8}
co(movies, field = "genres", sep = ",",
   similarity = "jaccard", counting = "fractional",
   scale = "minmax", min_occur = 15, threshold = 0.05,
   output = "cograph") |>
  splot(layout = "gephi", edge_width = 3, label_size = 0.9,
                 title = "IMDB Genre Co-occurrence (Jaccard, fractional, min 15 movies)")
```

## References

van Eck, N. J., & Waltman, L. (2009). How to normalize cooccurrence data? An analysis of some well‐known similarity measures. *Journal of the American Society for Information Science and Technology*, *60*(8), 1635-1651.

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/converters.R
\name{as_netobject}
\alias{as_netobject}
\alias{as_netobject.cooccurrence}
\title{Convert to Nestimate netobject}
\usage{
as_netobject(x, ...)

\method{as_netobject}{cooccurrence}(x, ...)
}
\arguments{
\item{x}{A \code{cooccurrence} data frame.}

\item{...}{Ignored.}
}
\value{
A \code{netobject} with class \code{c("netobject", "cograph_network")}.
}
\description{
Creates a \code{netobject} from a \code{cooccurrence} edge list,
compatible with \code{Nestimate::centrality()},
\code{Nestimate::bootstrap_network()}, etc.
}
\examples{
res <- cooccurrence(list(c("A","B","C"), c("B","C"), c("A","C")))
if (requireNamespace("Nestimate", quietly = TRUE)) {
  net <- as_netobject(res)
  net$n_nodes
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{actor_genres}
\alias{actor_genres}
\title{IMDB actor-genre long table (1970-2024)}
\format{
A data frame with 2,502 rows and 2 variables:
\describe{
\item{actor}{Actor name.}
\item{genre}{Genre label (one row per actor-genre combination).}
}
}
\source{
\url{https://developer.imdb.com/non-commercial-datasets/}
}
\usage{
actor_genres
}
\description{
Long-format table mapping each of the 624 actors in \code{\link{actors}}
to every genre of every movie they appeared in. Use this to build an
actor co-occurrence network grouped by genre: which actors share the
same genres? Pass \code{field = "actor"} and \code{by = "genre"} to
\code{\link{cooccurrence}}.
}
\examples{
head(actor_genres)
cooccurrence(actor_genres, field = "actor", by = "genre", similarity = "jaccard")
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/converters.R
\name{as_cograph}
\alias{as_cograph}
\alias{as_cograph.cooccurrence}
\title{Convert to cograph network}
\usage{
as_cograph(x, ...)

\method{as_cograph}{cooccurrence}(x, ...)
}
\arguments{
\item{x}{A \code{cooccurrence} data frame.}

\item{...}{Ignored.}
}
\value{
A \code{cograph_network} object.
}
\description{
Creates a \code{cograph_network} object from a \code{cooccurrence}
edge list, compatible with \code{cograph::splot()} and other cograph
functions.
}
\examples{
res <- cooccurrence(list(c("A","B","C"), c("B","C"), c("A","C")))
if (requireNamespace("cograph", quietly = TRUE)) {
  net <- as_cograph(res)
  net$n_nodes
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/launch_app.R
\name{launch_app}
\alias{launch_app}
\title{Launch the cooccure Shiny explorer}
\usage{
launch_app(...)
}
\arguments{
\item{...}{Passed to \code{\link[shiny]{runApp}} (e.g. \code{port},
\code{launch.browser}).}
}
\value{
Called for its side effect (launches the app). No return value.
}
\description{
Opens an interactive Shiny application for building and exploring
co-occurrence networks. Requires the \pkg{shiny} and \pkg{DT} packages.
}
\examples{
if (interactive()) {
  launch_app()
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/converters.R
\name{as_tidygraph}
\alias{as_tidygraph}
\alias{as_tidygraph.cooccurrence}
\title{Convert to tidygraph}
\usage{
as_tidygraph(x, ...)

\method{as_tidygraph}{cooccurrence}(x, ...)
}
\arguments{
\item{x}{A \code{cooccurrence} data frame.}

\item{...}{Ignored.}
}
\value{
A \code{tbl_graph} object.
}
\description{
Creates a \code{tbl_graph} from a \code{cooccurrence} edge list.
}
\examples{
res <- cooccurrence(list(c("A","B","C"), c("B","C"), c("A","C")))
if (requireNamespace("tidygraph", quietly = TRUE) &&
    requireNamespace("igraph",    quietly = TRUE)) {
  as_tidygraph(res)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/methods.R
\name{plot.cooccurrence}
\alias{plot.cooccurrence}
\title{Plot a cooccurrence network}
\usage{
\method{plot}{cooccurrence}(x, type = c("heatmap", "network"), ...)
}
\arguments{
\item{x}{A \code{cooccurrence} data frame.}

\item{type}{Character. \code{"heatmap"} (default) or \code{"network"}
(requires \pkg{igraph}).}

\item{...}{Passed to the plotting function.}
}
\value{
Invisibly returns \code{x}.
}
\description{
Plots the co-occurrence matrix as a heatmap. If \pkg{igraph} is available,
plots a network graph instead.
}
\examples{
res <- cooccurrence(list(c("A","B","C"), c("B","C"), c("A","C")))
plot(res)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/methods.R
\name{print.cooccurrence}
\alias{print.cooccurrence}
\title{Print a cooccurrence edge list}
\usage{
\method{print}{cooccurrence}(x, n = 10L, ...)
}
\arguments{
\item{x}{A \code{cooccurrence} data frame.}

\item{n}{Integer. Number of rows to show. Default 10.}

\item{...}{Ignored.}
}
\value{
Invisibly returns \code{x}.
}
\description{
Print a cooccurrence edge list
}
\examples{
res <- cooccurrence(list(c("A","B","C"), c("B","C"), c("A","C")))
print(res)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{actors}
\alias{actors}
\title{IMDB actor-movie long table (1970-2024)}
\format{
A data frame with 1,267 rows and 7 variables:
\describe{
\item{actor}{Actor name.}
\item{tconst}{IMDB title identifier linking to \code{\link{movies}}.}
\item{primaryTitle}{Movie title.}
\item{startYear}{Release year (integer).}
\item{decade}{Release decade as a character string.}
\item{genres}{Comma-separated genre labels for the linked movie.}
\item{averageRating}{IMDB average user rating for the linked movie.}
}
}
\source{
\url{https://developer.imdb.com/non-commercial-datasets/}
}
\usage{
actors
}
\description{
Long-format bipartite table linking actors to movies in
\code{\link{movies}}. Pre-filtered to the 624 actors who appear in at
least two movies, so all similarity measures compute instantly.
Pass \code{field = "actor"} and \code{by = "tconst"} to
\code{\link{cooccurrence}} to build an actor co-appearance network.
}
\examples{
head(actors)
cooccurrence(actors, field = "actor", by = "tconst", similarity = "jaccard")
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/converters.R
\name{as_igraph}
\alias{as_igraph}
\alias{as_igraph.cooccurrence}
\title{Convert to igraph}
\usage{
as_igraph(x, ...)

\method{as_igraph}{cooccurrence}(x, ...)
}
\arguments{
\item{x}{A \code{cooccurrence} data frame.}

\item{...}{Passed to \code{igraph::graph_from_data_frame}.}
}
\value{
An \code{igraph} object.
}
\description{
Creates an undirected, weighted \code{igraph} graph from a
\code{cooccurrence} edge list.
}
\examples{
res <- cooccurrence(list(c("A","B","C"), c("B","C"), c("A","C")))
if (requireNamespace("igraph", quietly = TRUE)) {
  g <- as_igraph(res)
  igraph::vcount(g)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/converters.R
\name{as_matrix}
\alias{as_matrix}
\alias{as_matrix.cooccurrence}
\title{Extract the co-occurrence matrix}
\usage{
as_matrix(x, ...)

\method{as_matrix}{cooccurrence}(x, type = c("normalized", "raw"), ...)
}
\arguments{
\item{x}{A \code{cooccurrence} data frame.}

\item{...}{Ignored.}

\item{type}{Character. \code{"normalized"} (default) or \code{"raw"}.}
}
\value{
A numeric matrix.
}
\description{
Returns the full square co-occurrence matrix (normalized + scaled).
Use \code{type = "raw"} for the raw count matrix.
}
\examples{
res <- cooccurrence(list(c("A","B","C"), c("B","C"), c("A","C")))
as_matrix(res)
as_matrix(res, type = "raw")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{movies}
\alias{movies}
\title{IMDB movie metadata (1970-2024)}
\format{
A data frame with 1,000 rows and 7 variables:
\describe{
\item{tconst}{IMDB title identifier (e.g. \code{"tt0068646"}).}
\item{primaryTitle}{Movie title.}
\item{startYear}{Release year (integer).}
\item{genres}{Comma-separated genre labels (e.g. \code{"Crime,Drama"}).}
\item{decade}{Release decade as a character string (e.g. \code{"1970s"}).}
\item{averageRating}{IMDB average user rating.}
\item{numVotes}{Number of IMDB user votes.}
}
}
\source{
\url{https://developer.imdb.com/non-commercial-datasets/}
}
\usage{
movies
}
\description{
A sample of 1,000 highly-rated IMDB movies (rating >= 7.0, >= 1,000 votes)
released between 1970 and 2024. The \code{genres} column is comma-delimited
and suitable for use as the \code{field} argument to \code{\link{cooccurrence}}.
}
\examples{
head(movies)
cooccurrence(movies, field = "genres", sep = ",", similarity = "jaccard")
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cooccurrence.R
\name{cooccurrence}
\alias{cooccurrence}
\alias{co}
\title{Build a co-occurrence network}
\usage{
cooccurrence(
  data,
  field = NULL,
  by = NULL,
  sep = NULL,
  weight_by = NULL,
  split_by = NULL,
  similarity = c("none", "jaccard", "cosine", "inclusion", "association", "dice",
    "equivalence", "relative"),
  counting = c("full", "fractional"),
  scale = NULL,
  threshold = 0,
  min_occur = 1L,
  top_n = NULL,
  output = c("default", "gephi", "igraph", "cograph", "matrix"),
  ...
)

co(
  data,
  field = NULL,
  by = NULL,
  sep = NULL,
  weight_by = NULL,
  split_by = NULL,
  similarity = c("none", "jaccard", "cosine", "inclusion", "association", "dice",
    "equivalence", "relative"),
  counting = c("full", "fractional"),
  scale = NULL,
  threshold = 0,
  min_occur = 1L,
  top_n = NULL,
  output = c("default", "gephi", "igraph", "cograph", "matrix"),
  ...
)
}
\arguments{
\item{data}{Input data. Accepts:
\itemize{
\item A \code{data.frame} with a delimited column (\code{field} + \code{sep}).
\item A \code{data.frame} in long/bipartite format (\code{field} + \code{by}).
\item A binary (0/1) \code{data.frame} or \code{matrix} (auto-detected).
\item A wide sequence \code{data.frame} or \code{matrix} (non-binary).
\item A \code{list} of character vectors (each element is a transaction).
}}

\item{field}{Character. The entity column --- determines what the nodes are.
For delimited format, a single column split by \code{sep}. For
long/bipartite, the item column. For multi-column delimited, a vector
of column names pooled per row. Use \code{field = "all"} for wide
sequence data (e.g. TraMineR / tna format) where every column is a
time point and cell values are the items.}

\item{by}{Character or \code{NULL}. Grouping column for long/bipartite
format. Each unique value defines one transaction.}

\item{sep}{Character or \code{NULL}. Separator for splitting delimited
fields.}

\item{weight_by}{Character or \code{NULL}. Column name containing a numeric
association strength for each entity-transaction pair. Only accepted for
long format (\code{field} + \code{by}). When supplied, each entity
contributes its weight rather than 1, so
\eqn{C_{ij} = \sum_d w_{id} \cdot w_{jd}}.
Typical use: topic-document probability matrices from LDA or similar
models.}

\item{split_by}{Character or \code{NULL}. Column name to split the data
by before computing co-occurrence. A separate network is computed per
group and the results are combined into a single data frame with an
additional \code{group} column. Only works with data.frame inputs.}

\item{similarity}{Character. Similarity measure:
\describe{
\item{\code{"none"}}{Raw co-occurrence counts.}
\item{\code{"jaccard"}}{\eqn{C_{ij} / (f_i + f_j - C_{ij})}.}
\item{\code{"cosine"}}{Salton's cosine:
\eqn{C_{ij} / \sqrt{f_i \cdot f_j}}.}
\item{\code{"inclusion"}}{Simpson coefficient:
\eqn{C_{ij} / \min(f_i, f_j)}.}
\item{\code{"association"}}{Association strength:
\eqn{C_{ij} / (f_i \cdot f_j)}
(van Eck & Waltman, 2009).}
\item{\code{"dice"}}{\eqn{2 C_{ij} / (f_i + f_j)}.}
\item{\code{"equivalence"}}{Salton's cosine squared:
\eqn{C_{ij}^2 / (f_i \cdot f_j)}.}
\item{\code{"relative"}}{Row-normalized: each row sums to 1.}
}}

\item{counting}{Character. Counting method:
\describe{
\item{\code{"full"}}{Each co-occurring pair adds 1 regardless of
transaction size. Default.}
\item{\code{"fractional"}}{Each pair adds \eqn{1 / (n_i - 1)}
where \eqn{n_i} is the number of items in transaction \eqn{i}.
Transactions with many items contribute less per pair
(Perianes-Rodriguez et al., 2016).}
}}

\item{scale}{Character or \code{NULL}. Optional scaling applied to weights
after similarity normalization:
\describe{
\item{\code{NULL} or \code{"none"}}{No scaling.}
\item{\code{"minmax"}}{Min-max to \eqn{[0, 1]}.}
\item{\code{"log"}}{Natural log: \eqn{\log(1 + w)}.}
\item{\code{"log10"}}{Log base 10: \eqn{\log_{10}(1 + w)}.}
\item{\code{"binary"}}{Binary: 1 if \eqn{w > 0}, else 0.}
\item{\code{"zscore"}}{Z-score standardization.}
\item{\code{"sqrt"}}{Square root.}
\item{\code{"proportion"}}{Divide by sum of all weights.}
}}

\item{threshold}{Numeric. Minimum edge weight to retain. Applied after
similarity and scaling. Default 0.}

\item{min_occur}{Integer. Minimum entity frequency. Entities appearing in
fewer than \code{min_occur} transactions are dropped. Default 1.}

\item{top_n}{Integer or \code{NULL}. Keep only the top \code{top_n} edges
by weight. When \code{split_by} is used, applied per group.
Default \code{NULL} (all edges).}

\item{output}{Character. Column naming convention for the output:
\describe{
\item{\code{"default"}}{\code{from}, \code{to}, \code{weight}, \code{count}.}
\item{\code{"gephi"}}{\code{Source}, \code{Target}, \code{Weight},
\code{Type} (= \code{"Undirected"}). Ready for Gephi import.}
\item{\code{"igraph"}}{Returns an \code{igraph} graph object directly.}
\item{\code{"cograph"}}{Returns a \code{cograph_network} object directly.}
\item{\code{"matrix"}}{Returns the square co-occurrence matrix.}
}}

\item{...}{Currently unused.}
}
\value{
Depends on \code{output}:
\itemize{
\item \code{"default"}: A \code{cooccurrence} data frame with columns
\code{from}, \code{to}, \code{weight}, \code{count} (and \code{group}
when \code{split_by} is used).
\item \code{"gephi"}: A data frame with columns \code{Source},
\code{Target}, \code{Weight}, \code{Type}, \code{Count}. Ready for
Gephi CSV import.
\item \code{"igraph"}: An \code{igraph} graph object.
\item \code{"cograph"}: A \code{cograph_network} object.
\item \code{"matrix"}: A square numeric co-occurrence matrix.
}
For the data frame outputs, rows are sorted by weight descending and
attributes store the full matrix, item frequencies, and parameters.
}
\description{
Constructs an undirected co-occurrence network from various input formats
and returns a tidy edge data frame. Argument names follow the citenets
convention.
}
\examples{
# Delimited keywords
df <- data.frame(
  id = 1:4,
  keywords = c("network; graph", "graph; matrix; network",
               "matrix; algebra", "network; algebra; graph")
)
cooccurrence(df, field = "keywords", sep = ";")

# Split by a grouping variable
df$year <- c(2020, 2020, 2021, 2021)
cooccurrence(df, field = "keywords", sep = ";", split_by = "year")

# List of transactions with Jaccard similarity
cooccurrence(list(c("A","B","C"), c("B","C"), c("A","C")),
             similarity = "jaccard")

# Short alias
co(df, field = "keywords", sep = ";", similarity = "cosine")

# Weighted long format (e.g. LDA topic-document probabilities)
theta <- data.frame(
  doc   = c("d1","d1","d1","d2","d2","d3","d3"),
  topic = c("T1","T2","T3","T1","T3","T2","T3"),
  prob  = c(0.6, 0.3, 0.1, 0.4, 0.6, 0.5, 0.5)
)
cooccurrence(theta, field = "topic", by = "doc", weight_by = "prob")

}
\references{
van Eck, N. J., & Waltman, L. (2009). How to normalize co-occurrence
data? An analysis of some well-known similarity measures. \emph{Journal of
the American Society for Information Science and Technology}, 60(8),
1635--1651.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/methods.R
\name{summary.cooccurrence}
\alias{summary.cooccurrence}
\title{Summarise a cooccurrence network}
\usage{
\method{summary}{cooccurrence}(object, ...)
}
\arguments{
\item{object}{A \code{cooccurrence} data frame.}

\item{...}{Ignored.}
}
\value{
Invisibly returns \code{object}.
}
\description{
Summarise a cooccurrence network
}
\examples{
res <- cooccurrence(list(c("A","B","C"), c("B","C"), c("A","C")))
summary(res)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{demo}
\alias{demo}
\title{Demo actor-movie-genre table}
\format{
A data frame with 34 rows and 3 variables:
\describe{
\item{movie}{Movie title.}
\item{actor}{Actor name.}
\item{genre}{Primary genre label.}
}
}
\usage{
demo
}
\description{
A small hand-crafted dataset of 30 well-known actors across 10 classic
films with genre labels. Designed for quick exploration in the Shiny app.
Use \code{field = "actor"} with \code{by = "movie"} or \code{by = "genre"}.
}
\examples{
head(demo)
cooccurrence(demo, field = "actor", by = "movie", similarity = "jaccard")
}
\keyword{datasets}
